#!/usr/bin/env python
#Copyright (c) 2017, University of Nebraska NIMBUS LAB  John-Paul Ore jore@cse.unl.edu
#All rights reserved.

#Redistribution and use in source and binary forms, with or without
#modification, are permitted provided that the following conditions are met:

#1. Redistributions of source code must retain the above copyright notice, this
   #list of conditions and the following disclaimer.
   #2. Redistributions in binary form must reproduce the above copyright notice,
      #this list of conditions and the following disclaimer in the documentation
         #and/or other materials provided with the distribution.

         #THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
         #ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
         #WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
         #DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
         #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
         #(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
         #LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
         #ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
         #(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
         #SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

         #The views and conclusions contained in the software and documentation are those
         #of the authors and should not be interpreted as representing official policies,
         #either expressed or implied, of the FreeBSD Project.

#           _          _ _                        _ _
#     _ __ | |__  _ __(_) | ___   _   _   _ _ __ (_) |_ ___
#    | '_ \| '_ \| '__| | |/ / | | | | | | | '_ \| | __/ __|
#    | |_) | | | | |  | |   <| |_| | | |_| | | | | | |_\__ \
#    | .__/|_| |_|_|  |_|_|\_\\__, |  \__,_|_| |_|_|\__|___/
#    |_|                      |___/


import sys
import os
import cppcheckdata                 # http://cppcheck.sourceforge.net/
from tree_walker import TreeWalker
import fileinput
import networkx as nx
from collections import OrderedDict
import inspect
from operator import attrgetter
from error_checker import ErrorChecker
from unit_error_types import UnitErrorTypes
# from database_connector import DatabaseConnector



class CPSUnitsChecker:
    ''' PHYSICAL UNITS CHECKING   
    '''

    def __init__(self):
        self.VERSION                    = 0.30
        self.debug                      = False
        self.debug_scope                = False
        self.debug_print_AST            = False
        self.debug_print_function_topo_sort = False
        self.debug_LHS                  = False
        self.debug_verbose              = False
        self.debug_configuration_count  = 0
        self.debug_function_count       = 0
        self.only_scan_these_functions  = []
        self.symbol_dict                = {}
        self.source_file                = ''
        self.errors                     = []  
        self.warnings                   = []
        self.error_checker              = None
        self.current_file_under_analysis= ''
        self.tw                         = None  # TREE WALKER
        self.source_file_lines          = []
        self.function_graph = nx.DiGraph()
        self.all_function_graphs        = []
        self.all_sorted_analysis_unit_dicts = []
        self.should_sort_by_function_graph = True
        self.count_of_files_containing_ros_units = 0
        self.found_ros_units_in_this_file = False
        self.all_tree_walkers           = []
        self.current_configuration      = None
        self.should_abandon_early       = True
        self.SHOULD_ONLY_FIND_FILES_WITH_UNITS = False
        self.SHOULD_WRITE_RESULTS_TO_DATABASE = False
        self.SHOULD_FIND_ALL_UNITS      = False
        self.SHOULD_MATCH_ON_HEURISTIC_VARIABLE_NAMES = False
        self.SHOULD_CHECK_UNIT_SMELLS   = False
        self.SHOULD_CHECK_MULTIPLE_UNITS= True
        self.SHOULD_CHECK_FUNCTION_ARGS = True
        self.SHOULD_CHECK_COMPARISONS   = True
        self.SHOULD_CHECK_LOGICAL_OPERATORS = True
        self.SHOULD_CHECK_ADDITION      = True  # DEFAULT TO TRUE
        self.CWD                        = '/Users/jore/courses/NIMBUS/RESEARCH/CPS_TYPES/cps_units'
        self.db_analysis_audit_row_id         = 0   # USED DURING INSERT
        self.current_repository         = ''   # USED DURING INSERT
        # http://stackoverflow.com/questions/14989858/get-the-current-git-hash-in-a-python-script
        self.git_version_command        = 'git describe --tags --always'  # MIGHT NEED THIS
        # INITIALIZE DATA STRUCTURE TO COUNT ERRORS BY TYPE
        self.list_of_error_counts_by_type = [0 for x in UnitErrorTypes().ERR_TYPE_NAMES]
        if self.debug_verbose:
            print inspect.stack()[0][3]
        self.all_classes_and_units_for_this_file_as_dict = {}
        # self.database_connector = None
        
    # def init_database_connection(self,
    # def init_database_connection(self):
        # """ CREATES THE OBJECT RESPONSIBLE FOR A CONNECTION TO A DATABASE FOR STORING RESULTS.
        # :returns: Nothing
        # : side-effects : creates database object attached to this object
        # """
        # self.database_connector = DatabaseConnector()


    def add_class_and_units_to_all_list(self, class_name, units):
        ''' WHEN TRACKING ALL UNIT TYPES, ADD THEM TO DATA STRUCTURE HERE
            input:  class_name str the name of the ROS class or str (ex: 'atan2') that merits units
                    units : the dictionary or units
            output: None, side effect updates local data structures.
            '''
        if not class_name in self.all_classes_and_units_for_this_file_as_dict:
            self.all_classes_and_units_for_this_file_as_dict[class_name] = units
        else:
            pass  # WE DON'T CASE BECAUSE WE'RE ONLY TRACKING UNIQUE UNITS, NOT COUNTS



    def handle_debug_verbose(self, message_string):
        print ("%d %d %s" % (self.debug_configuration_count,  self.debug_function_count, message_string))

    def handle_command_line_input(self): 
        for line in fileinput.input():
            print line

    def find_functions(self, a_cppcheck_configuration):
        ''' LINEAR SCAN THROUGH TOKENS TO FIND 'function' TOKENS THAT HAVE GLOBAL SCOPE.
            COLLECT AND RETURN DICT CONTAINING FUNCTION START AND END TOKEN POINTERS 

            returns: dict containing function start and end tokens
        '''
        function_dicts = {}

        if self.debug_verbose:
            print inspect.stack()[0][3]
        # FIND FUNCTIONS IN 'SCOPES' REGION OF DUMP FILE, START AND END TOKENs
        for s in a_cppcheck_configuration.scopes:
            if s.type=='Function': 
                # SCAN ALL FUNCTIONS UNLESS LIST OF FUNCTIONS SPECIFIED
                if self.only_scan_these_functions:
                    if  s.className not in self.only_scan_these_functions:
                        print 'cont...'
                        continue
                function_dicts[s.Id] = {'name': s.className,
                                        'linern': s.classStart.linenr,
                                        'tokenStart': s.classStart, 
                                        'tokenEnd': s.classEnd, 
                                        'scopeObject':s,
                                        'symbol_table':{},
                                        'function_graph_edges':[],
                                        'function':s.function}
                # CONSTRUCT LIST OF ROOT TOKENS
                function_dicts[s.Id]['root_tokens'] = self.find_root_tokens(s.classStart, s.classEnd)
                    
        if self.debug: 
            print "Found %d functions..." % len(function_dicts)
        return function_dicts


    def find_root_tokens(self, tokenStart, tokenEnd):
        ''' FOR A FUNCTION DEFIND AS ALL TOKENS FROM tokenStart TO tokenEnd, FIND THE ROOTS
            input: tokenStart  a CPPCheckData Token, first token in a function
            input: tokenEnd    a CPPCheckData Token, last token in a function

            output: a list of root_tokens, in flow order
            '''

        if self.debug_verbose:
            print inspect.stack()[0][3]
        root_tokens_set = set()
        current_token = tokenStart
        while(current_token != tokenEnd):  #todo: reverse token set exploration to top-down instead of bottom-up
            # HAS A PARENT
            if current_token.astParent: 
                a_parent = current_token.astParent
                has_parent = True
                while has_parent:
                    # HAS NO PARENT, THEREFORE IS ROOT
                    if not a_parent.astParent:     
                        root_tokens_set.add(a_parent)
                        a_parent.isRoot = True  # THIS PROPERTY IS A CUSTOM NEW PROPERTY
                        has_parent = False
                    else:
                        a_parent = a_parent.astParent 
            current_token = current_token.next

        
        root_tokens = list(root_tokens_set)
        # SORT NUMERICALLY BY LINE NUMBER
        root_tokens = sorted(root_tokens, key=lambda x : int(x.linenr))
        #for t in root_tokens:
            #print t.linenr
        return root_tokens


    def analyze_function(self, function_dict):
        
        if self.debug_verbose:
            print inspect.stack()[0][3]
        tw = TreeWalker(function_dict)  # ARGUMENT 'FUNCTION DICTIONARY' IS USED BY SYMBOL TABLE
        tw.cps_unit_checker = self
        tw.my_symbol_helper.cps_unit_checker = self
        self.tw = tw   # LINK CURRENT Tree Walker to self
        tw.debug = self.debug
        tw.debug_verbose = self.debug_verbose
        tw.current_file = self.current_file_under_analysis
        tw.source_file_lines = self.source_file_lines
        tw.source_file = self.source_file
        tw.should_check_unitless_during_multiplication = False # UNITLESS
        tw.my_symbol_helper.should_match_on_heuristic_variable_names = self.SHOULD_MATCH_ON_HEURISTIC_VARIABLE_NAMES
        tw.current_function_dict = function_dict

        found_units_in_this_function = False

        # ASSUME THE TOKENS COME BACK AS A SORTED LIST
        break_point = 1000
        i=0

        for root_token in function_dict['root_tokens']:
            # INITIALIZAE TREE WALKER FOR THIS ROOT TOKEN
            tw.was_some_unit_changed = True
            tw.is_unit_propagation_based_on_constants = False
            tw.is_unit_propagation_based_on_unknown_variable = False
            if self.SHOULD_ONLY_FIND_FILES_WITH_UNITS and self.found_ros_units_in_this_file and not self.SHOULD_FIND_ALL_UNITS:
                return
            # RESET THE TREE WALKER'S LINE NUMBERS
            tw.reset_min_max_line_numbers()
            # FIND THE MIN AND MAX LINE NUMBERS IN THIS AST : USED TO PROTECT LOOP FROM MULTI-LINE STATEMENTS
            tw.generic_recurse_and_apply_function(root_token, tw.find_min_max_line_numbers)
            
            # APPLY ROS UNITS TO VARIABLES
            tw.generic_recurse_and_apply_function(root_token, tw.apply_ROS_units)
            # APPLY UNITS FROM PREVIOUS ASSIGNMENTS IN SCOPE
            tw.generic_recurse_and_apply_function(root_token, tw.apply_units_from_scope)
            # APPLY UNITS FROM KNOWN FUNCTIONS - atan2
            tw.generic_recurse_and_apply_function(root_token, tw.apply_units_inverse_trig)
            # APPLY UNITS FROM KNOWN SYMBOLS - M_PI
            tw.generic_recurse_and_apply_function(root_token, tw.apply_units_known_symbols)
            # APPLY UNITS FROM KNOWN FUNCTION - toSec
            tw.generic_recurse_and_apply_function(root_token, tw.apply_units_toSec)
            # APPLY UNITS FROM KNOWN FUNCTION - getX, getY, getZ  
            tw.generic_recurse_and_apply_function(root_token, tw.apply_units_getXYZ)
            # APPLY UNITS FROM KNOWN FUNCTION - quatToRPY
            tw.generic_recurse_and_apply_function(root_token, tw.apply_units_quatToRPY)

            # CONTINUE TO ATTEMPT CHANGES UNTIL CHANGES CEASE
            while tw.was_some_unit_changed:  
                if i>break_point:
                    if self.debug_verbose:
                        s = "BREAKING QUIESSENCE (WHILE) LOOP AT %d" % break_point
                        raise ValueError(s)
                        return
                    break
                i+=1
                tw.was_some_unit_changed = False
                # LOOK FOR EARLY ABANDONMENT OF THIS AST
                if not tw.found_units_in_this_tree and self.should_abandon_early:
                    break
                # SPECIAL MODE - JUST FIND ROS FILES
                if self.SHOULD_ONLY_FIND_FILES_WITH_UNITS:
                    break
                ### PROPAGATE ACROSS '.'
                tw.generic_recurse_and_apply_function(root_token, tw.propagate_units_across_dot_connectors)
                ### PROPAGATE ACROSS + - * /  += -= *= /=
                tw.generic_recurse_and_apply_function(root_token, tw.propagate_units_across_operators)
                ### PROPAGATE ACROSS (
                tw.generic_recurse_and_apply_function(root_token, tw.propagate_units_across_parenthesis)
                ### PROPAGATE ACROSS [
                tw.generic_recurse_and_apply_function(root_token, tw.propagate_units_across_square_brackets)
                ### PROPAGATE INTO f(meters, arg2)  into function  - FUTURE: CONSTRAINT MODE
                #tw.generic_recurse_and_apply_function(root_token, tw.propagate_units_into_function_args)
                ### PROPAGATE ACROSS '='
                tw.generic_recurse_and_apply_function(root_token, tw.propagate_units_across_assignment)
                ### PROPAGATE ACROSS 'std::abs' 'std::fabs' 'abs' 'fabs'
                tw.generic_recurse_and_apply_function(root_token, tw.propagate_units_math_abs_fabs_floor_ceil)
                ### PROPAGATE ACROSS 'std::min' 'std::max' 'min' 'max'
                tw.generic_recurse_and_apply_function(root_token, tw.propagate_units_math_min_max)
                ### PROPAGATE ACROSS 'sqrt'
                tw.generic_recurse_and_apply_function(root_token, tw.propagate_units_sqrt)
                ### PROPAGATE ACROSS 'getXYZ
                # tw.generic_recurse_and_apply_function(root_token, tw.propagate_units_getXYZ)
                ### PROPAGATE ACROSS TERNARY
                tw.generic_recurse_and_apply_function(root_token, tw.propagate_units_ternary)
                ### PROPAGATE ACROSS 'pow'
                tw.generic_recurse_and_apply_function(root_token, tw.propagate_units_pow)
                ### PROPAGATE ACROSS 'atan2', 'acos', 'asin'
                tw.generic_recurse_and_apply_function(root_token, tw.propagate_units_inverse_trig)
                ### PROPAGATE ACROSS 'return'
                tw.generic_recurse_and_apply_function(root_token, tw.propagate_units_across_return)
                ### COLLECT UNITS FROM FUNCTION CALL POINTS
                tw.generic_recurse_and_apply_function(root_token, tw.collect_function_parameters_units_and_decorate_function_objects)
                 #FUNCTIONS todo:
            # END -- WHILE LOOP
            #ASSIGNMENT STATEMENTS WITH UNITS - RIGHT TO LEFT PROPAGATION
            if root_token.isAssignmentOp and root_token.units:   
                # COLLECT LEFT SIDE OF ASSIGNMENT
                lhs_compound_name = tw.recurse_and_collect_string(root_token.astOperand1)
                # ADD UNITS TO VARIABLE IN SCOPE
                tw.return_value_list = []
                tw.generic_recurse_and_apply_function(root_token.astOperand1, tw.find_first_variable)
                if len(tw.return_value_list) != 1:
                    if len(tw.return_value_list) == 0:
                        if self.debug and self.debug_LHS:
                            print "%s %s LHS no variable found not found" % (root_token.file, root_token.linenr)
                    else:
                        if self.debug and self.debug_LHS:
                            for value in tw.return_value_list:
                                print "%s %s LHS Variable found: %s" % (root_token.file, root_token.linenr, value)
                    continue # todo track count of this error
                lhs_var= tw.return_value_list[0]
                lhs_scope = lhs_var.nameToken.scope
                if lhs_var.isArgument:
                    lhs_scope = function_dict['scopeObject']
                        
                line_num = int(root_token.linenr)
                if lhs_compound_name in lhs_scope.var_ordered_dict:
                    # ALREADY HAS ASSIGNMENT, ADD TO IT 
                    lhs_scope.var_ordered_dict[lhs_compound_name][line_num] =  {
                                    'units':root_token.units,  # THIS IS A LIST 
                                    'token':root_token,
                                    'is_unit_propagation_based_on_constants':tw.is_unit_propagation_based_on_constants,
                                    'is_unit_propagation_based_on_unknown_variable':tw.is_unit_propagation_based_on_unknown_variable}
                else:
                    lhs_scope.var_ordered_dict[lhs_compound_name] = {
                            line_num : {'units':root_token.units, 
                                        'token':root_token,
                                        'is_unit_propagation_based_on_constants':tw.is_unit_propagation_based_on_constants,
                                        'is_unit_propagation_based_on_unknown_variable':tw.is_unit_propagation_based_on_unknown_variable}}
                                        
                if self.debug and self.debug_scope:
                    print "ADDED to SCOPE %s : %s has units %s at %s" % (lhs_scope.Id, lhs_compound_name, root_token.units, line_num)
                    print 'constants-based? %s  unknown-based? %s' % (tw.is_unit_propagation_based_on_constants,  tw.is_unit_propagation_based_on_unknown_variable)
            # EXTRACT UNITS FROM RETURN STATEMENT AND APPLY TO FUNCTION
            if root_token.str == 'return' and root_token.units:
                if function_dict['scopeObject'].function:
                    function_return_units = function_dict['scopeObject'].function.return_units
                    for u in root_token.units:
                        if u not in function_return_units:
                            function_return_units.append(u)
                            function_dict['scopeObject'].function.is_unit_propagation_based_on_constants = tw.is_unit_propagation_based_on_constants
                            function_dict['scopeObject'].function.is_unit_propagation_based_on_unknown_variable = tw.is_unit_propagation_based_on_unknown_variable
            found_units_in_this_function = found_units_in_this_function or tw.found_units_in_this_tree

        # LOOK FOR EARLY ABANDONMENT
        if not found_units_in_this_function and not self.debug_print_AST and self.should_abandon_early:
            return

        # DEBUG PRINTING
        for root_token in function_dict['root_tokens']:
            if self.debug_print_AST:
                tw.debug_walk(root_token)
 
        self.found_ros_units_in_this_file = self.found_ros_units_in_this_file or found_units_in_this_function
            

    def init_cppcheck_config_data_structures(self, cppcheck_configuration):  # initialize
        c = cppcheck_configuration
        for t in c.tokenlist:
            t.units = []
            t.is_unit_propagation_based_on_constants = False
            t.is_unit_propagation_based_on_unknown_variable = False
            t.isRoot = False
        for s in c.scopes:
            s.var_ordered_dict = OrderedDict()
        for f in c.functions:
            f.return_units = []
            f.arg_units = []
            f.is_unit_propagation_based_on_constants = False
            f.is_unit_propagation_based_on_unknown_variable = False
            for arg_number in f.argument.keys():
                f.arg_units.append([])
            f.arg_units_dict = {}
            for k,v in f.argumentId.iteritems():
                f.arg_units_dict[k] = []
        return c


    def init_cppcheckdata_structures(self, dump_file_as_cppcheckdata):
        '''  AUGMENT SCOPE DATA STRUCTURE WITH ADDITIONAL FEATURES TO SUPPORT THIS ANALYSIS
            each token gets an empty list to hold unit assignments
            each scope gets an empty list to hold variable unit assignments in that scope
            '''
        if self.debug_verbose:
            print inspect.stack()[0][3]
        for c in dump_file_as_cppcheckdata.configurations:
            c = self.init_cppcheck_config_data_structures(c)
                    
        return dump_file_as_cppcheckdata


    def main_run_check(self, dump_file, source_file=''): 
        ''' PEFORM UNITS CHECKING
            todo: pep8
            input: a cppcheck 'dump' file containing an Abstract Syntax Tree (AST), symbol table, and token list.
            returns: None
            side-effects: updates datbase with information about this unit analysis
        '''
        if self.debug_verbose:
            print inspect.stack()[0][3]
        self.source_file = source_file
        self.current_file_under_analysis = dump_file
        # PARSE INPUT
        data = cppcheckdata.parsedump(dump_file)
        analysis_unit_dict = {}
        # INITIALIZE ERROR CHECKING OBJECT
        self.error_checker = ErrorChecker(self.debug, source_file, self)
        # GIVE TREE WALKER ACCESS TO SOURCE FILE FOR DEBUG PRINT
        if self.source_file and self.debug and os.path.exists(self.source_file):
            with open(self.source_file) as f:
                self.source_file_lines = f.readlines()
                print "yes"
        else:
            if self.debug:
                print "no %s %s" % (self.source_file, self.debug)
        # for c in data.configurations:  #todo: what is a data configuration?  -- Check for multiple
        for c in data.configurations[:1]:   # MODIFIED TO ONLY TEST THE FIRST CONFIGURATION
            self.current_configuration = c
        # ADD AST DECORATION PLACEHOLDERS
            c = self.init_cppcheck_config_data_structures(c)
            # REFRESH VARIABLES
            self.function_graph = nx.DiGraph()
            # GET DICT OF ALL GLOBALLY SCOPED FUNCTIONS
            analysis_unit_dict = self.find_functions(c)
            sorted_analysis_unit_dict = analysis_unit_dict;  # WILL BECOME AN ORDERED DICT IF self.should_sort_by_function_graph
            # FIND ORDER FOR FUNCTION GRAPH EXPLORATION ( topo sort, if possible, otherwise todo ??)
            if self.should_sort_by_function_graph:
                self.build_function_graph(analysis_unit_dict)  # WILL USE DAG SUBGRAPH
                sorted_analysis_unit_dict = self.make_sorted_analysis_unit_dict_from_function_graph(analysis_unit_dict) # RETURNS ORDERED DICT
                self.all_sorted_analysis_unit_dicts.append(sorted_analysis_unit_dict)
                # SPECIAL MODE FOR COUNTING / IDENTIFYING FILES WITH ROS UNITS
                if self.SHOULD_ONLY_FIND_FILES_WITH_UNITS and self.found_ros_units_in_this_file:
                    return
            ##print "G nodes:%d edges:%d" % (G.order(), G.size())
            if self.debug_print_function_topo_sort:
                self.debug_print_function_graph(sorted_analysis_unit_dict)
            # COLLECT ALL TOKEN PARSE TREES FOR EACH FUNCTION
            for function_dict in sorted_analysis_unit_dict.values():
                self.analyze_function(function_dict)
                # SAVE OFF A POINTER TO THE TREE WALKERS USED IN THIS ANALYSIS FOR TESTING
                self.all_tree_walkers.append(self.tw)
                self.debug_function_count += 1
            # if self.debug:
                # for e in self.errors:
                    # print e['error_msg']
            # self.error_checker.error_check_function_args_consistent(c)
            # for e in self.error_checker.all_errors:
                # self.errors.append(e)
            # ADD ERROR OBJECTS FOUND BY ERROR CHECKER TO MY ERRORS
            if self.debug :
                self.error_checker.pretty_print()
            # DEBUG COUNTERS
            self.debug_configuration_count += 1
            self.debug_function_count = 0
        ## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
        ## OPTIONAL FIND UNITS ONLY AND WRITE TO DATABASE
        ## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
        # if self.SHOULD_ONLY_FIND_FILES_WITH_UNITS and self.SHOULD_FIND_ALL_UNITS:
            # self.insert_file_unit_class_records()
            # return
        if self.SHOULD_ONLY_FIND_FILES_WITH_UNITS and self.found_ros_units_in_this_file:
            with open('2016_12_23_github_files_with_units.txt', 'a') as f:
                f.write(self.current_file_under_analysis + '\n')
        ## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
        ## MAIN ERROR CHECKING
        ## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
        # for c in data.configurations: 
        for c in data.configurations[:1]:  # FIRST ONLY
            # CHECK THIS CONFIGURATION FOR ERRORS
            if self.SHOULD_CHECK_MULTIPLE_UNITS:
                self.error_checker.error_check_multiple_units(c)
            if self.SHOULD_CHECK_FUNCTION_ARGS:
                self.error_checker.error_check_function_args_consistent(c)
            if self.SHOULD_CHECK_UNIT_SMELLS:
                self.error_checker.error_check_unit_smell(c)
            # todo: when the same error occurs in multiple data configs, where is a good place to catch that
            # todo: add checking of addition on RH
        for sorted_analysis_unit_dict in self.all_sorted_analysis_unit_dicts:
            if self.SHOULD_CHECK_COMPARISONS:
                self.error_checker.error_check_comparisons(sorted_analysis_unit_dict)
            if self.SHOULD_CHECK_LOGICAL_OPERATORS:
                self.error_checker.error_check_logical_operators(sorted_analysis_unit_dict)
            if self.SHOULD_CHECK_ADDITION:
                self.error_checker.error_check_addition_of_incompatible_units(sorted_analysis_unit_dict)
        # GET ERRORS FROM ERROR CHECKER OBJECT  (add to list of pointers)
        for e in self.error_checker.all_errors:
            self.errors.append(e)
            # COUNTS ERRORS BY TYPE
            # self.list_of_error_counts_by_type[e.ERROR_TYPE] += 1
        # if self.SHOULD_WRITE_RESULTS_TO_DATABASE:
            # if not self.database_connector:
                # print ("Warning: no database connection object available--not writing results to database.")
            # else:
                # self.database_connector.found_ros_units_in_this_file = self.found_ros_units_in_this_file
                # if self.errors:
                    # self.database_connector.update_database(self.errors)


    def build_function_graph(self, analysis_unit_dict):
        ''' BUILDS DIRECTED FUNCTION GRAPH 
            input:  a dictionary of functions from this dump file
            output: none.  Side effect creates a graph linked to this object
            '''
        if self.debug_verbose:
            print inspect.stack()[0][3]
        # BUILD CALL GRAPH
        self.function_graph = nx.DiGraph()
        G = self.function_graph
        for k, function_dict in analysis_unit_dict.iteritems():
            if function_dict['function']:  # MIGHT BE NONE WHEN FUNCTION IS CLASS CONSTRUCTOR (?)
                node = function_dict['function'].Id             # Id of the Function
                #if not G.has_node(node):
                G.add_node(node) #, function_dict_key=k})  # FUNCTION CPP OBJECT IS NODE
                #else:
                all_attr = nx.get_node_attributes(G, 'function_id')
                all_attr[node] = k
                nx.set_node_attributes(G, 'function_id', all_attr)
                #function_dict['is_visited'] = False
                self.add_edges_to_function_graph(function_dict, G, node)
        self.function_graph = G


    def make_sorted_analysis_unit_dict_from_function_graph(self, analysis_unit_dict):
        ''' BUILDS A TOPO SORTED FUNCTION GRAPH.  THIS ALLOWS THE ANALYSIS TO START ON FUNCTION LEAFS, 
            SO WE CAN HOPEFULLY DISCOVER UNITS ON THE RETURN TYPE AND PROPAGE THEM UP.  THE 
            FUNCTION GRAPH MAY HAVE CYCLES (recursion, for example), THEREFORE WE REMOVE THESE EDGES FROM THE GRAPH
            AND ANALYZE THEM LAST (<-- not sure this is best)
            input:  a dictionary of functions from this dump file
            output: OrderedDict of functions
            postcondition:   returned dict must be the same length as the input dict, and contain all the same elements
            '''
        if self.debug_verbose:
            print inspect.stack()[0][3]
        return_dict = OrderedDict()
        G = self.function_graph 
        # TRY FINDING A DAG.  IF NOT, REMOVE EDGES AND TRY AGAIN. 
        super_break = 0
        while nx.number_of_nodes(G) > 0 and super_break < 1000:
            super_break +=1 
            if not nx.is_directed_acyclic_graph(G):
                try:
                    # SEARCH FOR CYCLE AND REMOVE EDGES
                    edges = nx.find_cycle(G)
                    G.remove_edges_from(edges)
                    if self.debug and self.debug_print_function_graph:
                        print 'Function graph has cycle %s' % edges,
                except:
                    if self.debug and self.debug_print_function_graph:
                        print 'Function graph is not a DAG and does not have a cycle!'
                        # GIVE UP AND RETURN UNSORTED 
                    return analysis_unit_dict
            else:
                # WE HAVE A DIGRAPH, CAN PROCEED ( and topo sort )
                break
        
        if nx.number_of_nodes(G) == 0:
            # RETURN UNCHANGED
            return analysis_unit_dict
        # WE HAVE A DIRECTED GRAPH WITH NODES, CAN SORT AND ADD NODES TO ORDERED LIST
        function_graph_topo_sort = nx.topological_sort(G)
        function_graph_topo_sort_reversed = function_graph_topo_sort[::-1]  # REVERSED
        
        # OPTIONAL DEBUG PRINT
        if self.debug_print_function_topo_sort:
            print function_graph_topo_sort_reversed


        # CREATE RETURN DICT FROM TOPO SORT
        for node in function_graph_topo_sort_reversed:
            function_id_attr_dict = nx.get_node_attributes(G, 'function_id')
            #print function_id_attr_dict
            if node in function_id_attr_dict:  # FIRST OPTION SHOULD SHORTCUT WHEN ATTR DICT DOES NOT EXITS
                # ADD FUNCTION TO NEW DICTIONARY - THIS IS THE EXPLORE ORDER
                return_dict[function_id_attr_dict[node]] = analysis_unit_dict[function_id_attr_dict[node]]
            else:
                if self.debug and self.debug_print_function_graph:
                    print "Graph node does not have function_dict_key"
        
        # ADD ANY REMAINING FUNCTIONS NOT IN THE TOPO SORT TO THE ORDERED DICT
        for k in analysis_unit_dict.keys():
            if k not in return_dict:
                return_dict[k] = analysis_unit_dict[k]

        assert (len(return_dict) == len(analysis_unit_dict))

        return return_dict


    def debug_print_function_graph(self, analysis_unit_dict):
        if self.debug_verbose:
            print inspect.stack()[0][3]
        if not analysis_unit_dict:
            return
        for function_dict in analysis_unit_dict.values():
            print "%s :" % function_dict['name']
            for edge in function_dict['function_graph_edges']:
                print ' --> %s' % edge.name
            

    def add_edges_to_function_graph(self, function_dict, G, current_node):
        if self.debug_verbose:
            print inspect.stack()[0][3]
            #self.print 'cps_unit_checker: add_edges_to_function_graph'
        
        # GET THE FIRST TOKEN AFTER THE FUNCTION DEFINITION
        current_token = function_dict['tokenStart']
        end_token = function_dict['tokenEnd']
        while current_token is not end_token: # TERMINATION GUARENTEED IF DUMP FILE IS WELL FORMED
            current_token = current_token.next  # ON FIRST LOOP, SKIPS SELF-REFERENCE 
            if current_token.function:
                if not G.has_edge(current_node, current_token.function.Id):
                    G.add_edge(current_node, current_token.function.Id)
      


if __name__ == "__main__":
    # FOR COMMAND LINE INPUT
    cps_unit_checker = CPSUnitsChecker()
    cps_unit_checker.handle_command_line_input()







from symbol_helper import SymbolHelper
import copy
import inspect


class TreeWalker:

    name = None

    def __init__(self, function_dictionary):
        self.my_symbol_helper = SymbolHelper(function_dictionary)
        self.symbol_helper = self.my_symbol_helper
        self.debug = False
        self.debug_scope = False
        self.debug_verbose = False
        self.source_file = ''
        self.source_file_lines = []
        self.errors = []
        self.return_value = None
        self.return_value_list = []
        self.current_file = ''
        self.current_function_dict = None
        self.should_check_unitless_during_multiplication = False
        self.found_units_in_this_tree = False
        self.was_some_unit_changed = False
        self.cps_unit_checker = None   # DEPENDENCY INJECTION.  SEE MARTIN FOWLER
        self.is_unit_propagation_based_on_unknown_variable = False
        self.is_unit_propagation_based_on_constants = False
        self.is_assignment_statement = False
        self.current_AST_start_line_number = None  # PROTECTS FROM MULTI-LINE STATEMENTS
        self.current_AST_end_line_number = None  # PROTECTS FROM MULTI-LINE STATEMENTS
        pass

    def generic_recurse_and_apply_function(self, token, function_to_apply):
        ''' GENERIC RECURSION PATTERN - LEFT RIGHT TOKEN
            input:  token  CPPCHECK token to recurse upon
            function_to_apply:  the function to apply to the token 
                after recursing.
            returns: None   Side effect determined by function_to_apply
            '''
        if not token:
            return
        # INITIALIZE
        left_token = right_token = None

        # LEFT
        if token.astOperand1:
            left_token = token.astOperand1
            self.generic_recurse_and_apply_function(left_token,
                                                    function_to_apply)
        # RIGHT
        if token.astOperand2:
            right_token = token.astOperand2
            self.generic_recurse_and_apply_function(right_token,
                                                    function_to_apply)

        function_to_apply(token, left_token, right_token)

    def find_assignment_tokens_recursive_target(self,
                                                token,
                                                left_token,
                                                right_token):
        ''' FIND IF THIS AST IS AN ASSIGNMENT STATEMENT
            input:  token AN AST TOKEN
                    left_token  (ignored)
                    right_token  (ignored)
            returns: None  (side effect: modifies class is_assignment_statement
            '''
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        if token.isAssignmentOp:
            self.is_assignment_statement = True

    def find_min_max_line_numbers(self, token, left_token, right_token):
        ''' FIND THE MIN AND MAX LINE NUMBERS FOR THIS AST,
                PROTECT FROM MULTI-LINE STATEMENTS
            input:  token AN AST TOKEN
                    left_token  (ignored)
                    right_token  (ignored)
            returns: None  (side effect: modifies class min and max
                line number range
            '''
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        if not self.current_AST_start_line_number or \
                token.linenr < self.current_AST_start_line_number:
            self.current_AST_start_line_number = token.linenr
        if not self.current_AST_end_line_number or \
                token.linenr > self.current_AST_end_line_number:
            self.current_AST_end_line_number = token.linenr

    def reset_min_max_line_numbers(self):
        ''' INITIALIZES THE AST LINE NUMBERS BACK TO NONE.
                SHOULD BE CALLED BEFORE EVERY AST IS EVALUATED
            input: None
            output: None.  side effect is setting class variables
                for min max to None
            '''
        self.current_AST_start_line_number = None
        self.current_AST_end_line_number = None

    def apply_ROS_units(self, token, left_token, right_token):
        '''  DECORATE LEAF WITH ROS MESSAGE TYPE UNITS
            input:  token AN AST TOKEN
                    left_token  (ignored)
                    right_token  (ignored)
            returns: None  (side effect: adds units to ROS variable leafs)
        '''
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        if token.variable:
            # LOOKUP VARIABLE TYPE IN SYMBOL DICTIONARY FOR ROS MESSAGES
            units_as_dict = self.my_symbol_helper.find_units_for_variable(token)
            # ADD THIS DICTIONARY
            if units_as_dict and (units_as_dict not in token.units):
                self.was_some_unit_changed = True
                # CHECK SYMBOL HELPER FOR WEAK INFERENCE IN THE CASE OF
                # JOINT_STATES AND getXYZ()
                if self.my_symbol_helper.is_weak_inference:
                    self.is_unit_propagation_based_on_unknown_variable = True
                    token.is_unit_propagation_based_on_unknown_variable = True
                if self.debug_verbose:
                    s = "tw. units changed in %s" % inspect.stack()[0][3]
                    print s
                token.units.append(units_as_dict)
                self.found_units_in_this_tree = True

    def apply_units_from_scope(self, token, left_token, right_token):
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        if token.variable:
            compound_var_name = self.my_symbol_helper.\
                    find_compound_variable_name_for_variable_token(token)
            # PATH TO THE SCOPE WHERE THIS VAR DEFINED
            var_scope = token.variable.nameToken.scope
            if token.variable.isArgument:
                if self.current_function_dict:
                    var_scope = self.current_function_dict['scopeObject']
            if self.debug and self.debug_scope:
                s = "%s attempting to apply previous units for %s "
                s += "to scope dict %s"
                s = s % (token.linenr,
                         compound_var_name,
                         var_scope.var_ordered_dict)
                pass
            if compound_var_name in var_scope.var_ordered_dict:
                # GET MOST RECENT ASSIGNMENT IN SCOPE
                #    SORT BY LINE NUMBER - WE WANT LAST (MOST RECENT)
                #    ASSIGNMENT todo: flow sensitive?
                # IF len(dict)==1, ASSUME SORT IS FAST
                s = sorted(var_scope.var_ordered_dict[compound_var_name])
                last_line_number = s[-1]
                # REJECT UNITS FOUND IN SCOPE IF THEY CAME FROM THIS AST
                # todo: potential BUG when previous assignment was from a different file but the same line number
                if last_line_number in range(int(self.current_AST_start_line_number), int(self.current_AST_end_line_number)+1):
                    return

                most_recent_units = var_scope.var_ordered_dict[compound_var_name][last_line_number]['units']
                for u in most_recent_units:
                    if u in token.units:
                        continue
                    token.units.append(u)
                    self.was_some_unit_changed = True
                    self.found_units_in_this_tree = True
                    # IF PREVIOUS ASSIGNMENT WAS BASED ON CONSTANTS OR UNKNOWN VARIABLES, PAY ATTENTION TO IT HERE
                    if var_scope.var_ordered_dict[compound_var_name][last_line_number]['is_unit_propagation_based_on_constants']:
                        self.is_unit_propagation_based_on_constants = True
                        token.is_unit_propagation_based_on_constants = True
                    if var_scope.var_ordered_dict[compound_var_name][last_line_number]['is_unit_propagation_based_on_unknown_variable']:
                        self.is_unit_propagation_based_on_unknown_variable = True
                        token.is_unit_propagation_based_on_unknown_variable = True
                    if self.debug_verbose:
                        s = "tw. units changed in %s" \
                                % inspect.stack()[0][3]
                        print s % inspect.stack()[0][3]
                        s = "len most recent %d"
                        print s % len(most_recent_units)
                        s =  "len most recent %d"
                        print s % len(most_recent_units)
                        s = "FOUND UNITS for %s on line %s from line "
                        s += "%s. %s"
                        print s % (compound_var_name,
                                   token.linenr,
                                   last_line_number,
                                   most_recent_units)
                        print "FOUND UNITS for %s on line %s from line %s. %s" % (compound_var_name, token.linenr, last_line_number, most_recent_units)
                    if self.debug and self.debug_scope:
                        print "FOUND UNITS for %s on line %s from line %s. %s" % (compound_var_name, token.linenr, last_line_number, most_recent_units)

                else:
                    pass # DO NOTHING BECAUSE CURRENT == MOST RECENT UNITS
        if token.function:
            function_return_units = token.function.return_units
            for u in function_return_units:
                if u not in token.units:
                    self.was_some_unit_changed = True
                    self.found_units_in_this_tree = True
                    if self.debug_verbose:
                        s = "tw. units changed in %s" % inspect.stack()[0][3]
                        print s
                    token.units.append(u)
                    self.is_unit_propagation_based_on_constants = \
                            token.function.\
                            is_unit_propagation_based_on_constants
                    token.is_unit_propagation_based_on_constants = \
                            token.function.\
                            is_unit_propagation_based_on_constants
                    self.is_unit_propagation_based_on_unknown_variable = \
                            token.function.\
                            is_unit_propagation_based_on_unknown_variable
                    token.is_unit_propagation_based_on_unknown_variable = \
                            token.function.\
                            is_unit_propagation_based_on_unknown_variable

    def apply_units_known_symbols(self, token, left_token, right_token):
        ''' INCLUDES M_PI
            '''
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        if token.str == 'M_PI':
            if token.str in self.my_symbol_helper.ros_unit_dictionary:
                a_dict = self.my_symbol_helper.ros_unit_dictionary\
                        [token.str][token.str]
                if a_dict not in token.units:
                    self.was_some_unit_changed = True
                    self.found_units_in_this_tree = True
                    if self.debug_verbose:
                        s = "tw. units changed in %s" % inspect.stack()[0][3]
                        print s
                    if self.cps_unit_checker:
                        if self.cps_unit_checker.SHOULD_FIND_ALL_UNITS:
                            self.cps_unit_checker.\
                                    add_class_and_units_to_all_list\
                                    (token.str, a_dict)
                    token.units.append(a_dict)

    def apply_units_inverse_trig(self, token, left_token, right_token):
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        if token.str in ['atan2', 'acos', 'asin', 'atan']:
            if token.str in self.my_symbol_helper.ros_unit_dictionary:
                a_dict = self.my_symbol_helper.ros_unit_dictionary[
                         token.str][token.str]
                if a_dict not in token.units:
                    self.was_some_unit_changed = True
                    self.found_units_in_this_tree = True
                    if self.debug_verbose:
                        s = "tw. units changed in %s" % inspect.stack()[0][3]
                        print s
                    if self.cps_unit_checker:
                        if self.cps_unit_checker.SHOULD_FIND_ALL_UNITS:
                            self.cps_unit_checker.\
                                    add_class_and_units_to_all_list\
                                    (token.str, a_dict)
                    token.units.append(a_dict)

    def apply_units_toSec(self, token, left_token, right_token):
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        if token.str == 'toSec':
            if token.str in self.my_symbol_helper.ros_unit_dictionary:
                toSec_units_dict = self.\
                        my_symbol_helper.ros_unit_dictionary[
                         token.str][token.str]
                if  toSec_units_dict not in token.units:
                    self.was_some_unit_changed = True
                    self.found_units_in_this_tree = True
                    if self.debug_verbose:
                        s = "tw. units changed in %s" % inspect.stack()[0][3]
                        print s
                    if self.cps_unit_checker:
                        if self.cps_unit_checker.SHOULD_FIND_ALL_UNITS:
                            self.cps_unit_checker.\
                                    add_class_and_units_to_all_listi\
                                    (token.str, toSec_units_dict)
                    token.units.append(toSec_units_dict)

    def apply_units_getXYZ(self, token, left_token, right_token):
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        if token.str not in ['getX', 'getY', 'getZ']:
            return
        # DECLARE VARIABLE HERE FOR SCOPING
        units_for_getXYZ = []
        # TRY TO FIGURE OUT TYPE THIS WAS CALLED ON
        # ASSUME PARENT TOKEN IS '.'
        # OPTION 1 - LOOK FOR getOrigin and getRotation
        if token.astParent.astOperand1 and \
                token.astParent.astOperand1.str == '(':
            paren_token = token.astParent.astOperand1  # CONVENIENCE
            if paren_token.astOperand1 and paren_token.astOperand1.str == '.':
                dot_token = paren_token.astOperand1

                if dot_token.astOperand1 and dot_token.astOperand1.variable:
                    class_name = self.my_symbol_helper.\
                            find_variable_type(dot_token.astOperand1.variable)
                    class_name_sanitized = self.my_symbol_helper.\
                            sanitize_class_name(class_name)
                    if class_name_sanitized in self.my_symbol_helper.\
                            ros_unit_dictionary:
                        if dot_token.astOperand2 and \
                                dot_token.astOperand2.str == 'getOrigin':
                            units_for_getXYZ = self.my_symbol_helper.\
                                    ros_unit_dictionary[
                                     class_name_sanitized]['getOrigin']
                        elif dot_token.astOperand2 and \
                                dot_token.astOperand2.str == 'getRotation':
                            units_for_getXYZ = self.\
                                    my_symbol_helper.ros_unit_dictionary[
                                     class_name_sanitized]['getRotation']
                else:
                    s = 'Failed token: %s %s' 
                    print s % (dot_token.astOperand1.linenr, 
                               dot_token.astOperand1.str)
        elif token.astParent.astOperand1:
            # GET TYPE SO WE CAN INFER CORRECT UNITS BASED
            # ON KNOWLEDGE OF ROS UNIT ASSUMPTIONS
            if token.astParent.astOperand1.variable:
                class_name = self.my_symbol_helper.\
                        find_variable_type(token.astParent.
                        astOperand1.variable)
                class_name_sanitized = self.my_symbol_helper.\
                        sanitize_class_name(class_name)
                # FIND APPROPRIATE UNITS
                if class_name_sanitized in \
                        self.my_symbol_helper.ros_unit_dictionary:
                    units_for_getXYZ = self.my_symbol_helper.\
                            ros_unit_dictionary[
                             class_name_sanitized][token.str]
            else:
                pass

        else:
            # NO AST PARENT OPERAND1 - SHOULD BE IMPOSSIBLE
            # assert False
            pass

        if units_for_getXYZ and (units_for_getXYZ not in token.units):
            self.was_some_unit_changed = True
            self.found_units_in_this_tree = True
            if self.debug_verbose:
                s = "tw. units changed in %s" % inspect.stack()[0][3]
                print s
            if self.cps_unit_checker:
                if self.cps_unit_checker.SHOULD_FIND_ALL_UNITS:
                    self.cps_unit_checker.\
                            add_class_and_units_to_all_list(
                            token.str,
                            units_for_getXYZ)
            token.units.append(units_for_getXYZ)

    def apply_units_quatToRPY(self, token, left_token, right_token):
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        if token.str not in ['quatToRPY']:
            return
        if token.str not in self.my_symbol_helper.ros_unit_dictionary:
            return
        rpy_tokens_list = []
        if token.astParent.astOperand2.str == ',':
            try:
                # THIS MAKES A LOT OF ASSUMPTIONS ABOUT THE AST STRUCTURE
                roll_token = token.astParent.astOperand2.\
                        astOperand1.astOperand1.astOperand2
                pitch_token = token.astParent.astOperand2.\
                        astOperand1.astOperand2
                yaw_token = token.astParent.astOperand2.astOperand2
                # ADD THESE TOKENS TO A LIST
                rpy_tokens_list = [roll_token, pitch_token, yaw_token]
            except:
                if self.debug:
                    print "quatToRPY 1st Case exception"
        elif token.astParent.str == "::":
            try:
                roll_token = token.astParent.astParent.astOperand2.\
                        astOperand1.astOperand1.astOperand2
                pitch_token = token.astParent.astParent.astOperand2.\
                        astOperand1.astOperand2
                yaw_token = token.astParent.astParent.astOperand2.astOperand2
                # ADD THESE TOKENS TO A LIST
                rpy_tokens_list = [roll_token, pitch_token, yaw_token]
            except:
                if self.debug:
                    print "quatToRPY 2nd Case exception"
        if not rpy_tokens_list:
            if self.debug:
                print "quatToRPY no rpy_tokens_list tokens"
            return
        # GET UNITS FROM SYMBOL HELPER
        quatToRPY_units_dict = self.my_symbol_helper.\
                ros_unit_dictionary[token.str][token.str]
        for rpy_token in rpy_tokens_list:
            rpy_compound_name = self.recurse_and_collect_string(rpy_token)
            # ADD UNITS TO VARIABLE IN SCOPE
            self.return_value_list = []
            self.generic_recurse_and_apply_function(rpy_token,
                                                    self.find_first_variable)
            if len(self.return_value_list) != 1:
                if self.debug:
                    s = "quatToRPY exception: multiple variables "
                    s += "returned in find_first_variable"
                    print s
                return  # todo: track count of this error
            rpy_var = self.return_value_list[0]
            rpy_scope = rpy_var.nameToken.scope
            linenr = int(rpy_token.linenr)
            if rpy_compound_name in rpy_scope.var_ordered_dict:
                # ALREADY HAS ASSIGNMENT, ADD TO IT
                rpy_scope.var_ordered_dict[rpy_compound_name][linenr]\
                        = {'units': [quatToRPY_units_dict],
                           'token': token,  # THIS IS A LIST
                           'is_unit_propagation_based_on_constants': False,
                           'is_unit_propagation_based_on_unknown_variable':
                                   False}
            else:
                rpy_scope.var_ordered_dict[rpy_compound_name]\
                        = {linenr:
                           {'units': [quatToRPY_units_dict],
                           'token': token,
                           'is_unit_propagation_based_on_constants': False,
                           'is_unit_propagation_based_on_unknown_variable':
                                   False}}

    def propagate_units_across_dot_connectors(self,
                                              token,
                                              left_token,
                                              right_token):
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        self.propagate_units_across_connectors(token,
                                               left_token,
                                               right_token,
                                               '.')
        # DOT CONNECTOR AFTER PAREN WiTH NOTHING BEFORE
        if token and token.str == '.' and \
                token.astParent and \
                token.astParent.str == '(' and not \
                token.astParent.astOperand2:
            self.propagate_units_across_connectors(token.astParent,
                                                   token,
                                                   None,
                                                   '(')

    def propagate_units_across_parenthesis(self,
                                           token,
                                           left_token,
                                           right_token):
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        if token.str == '(':
            if left_token.function:
                # IF LEFT SIDE IS A FUNCTION, PROPATE ITS RETURN UNITS
                for u in left_token.units:
                    if u not in token.units:
                        self.was_some_unit_changed = True
                        if self.debug_verbose:
                            s = "tw. units changed in %s"
                            s = s % inspect.stack()[0][3]
                            print s
                        token.units.append(u)
            elif left_token.isName:
                return
            elif not right_token:
                self.propagate_units_across_connectors(token,
                                                       left_token,
                                                       right_token,
                                                       '(')

            else:
                pass

    def propagate_units_across_square_brackets(self,
                                               token,
                                               left_token,
                                               right_token):
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        self.propagate_units_across_connectors(token,
                                               left_token,
                                               right_token,
                                               '[')

    def propagate_units_across_assignment(self,
                                          token,
                                          left_token,
                                          right_token):
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        # PREVENT PROPAGATION ACROSS MULTIPLE INITIALIZATIONS, LIKE
        # x = y = z = 0
        if token.str == "=":
            if left_token and left_token.str == "=":
                return
            if right_token and right_token.str == "=":
                return
        self.propagate_units_across_connectors(token,
                                               left_token,
                                               right_token,
                                               '=')

    def propagate_units_across_return(self, token, left_token, right_token):
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        self.propagate_units_across_connectors(token,
                                               left_token,
                                               right_token,
                                               'return')

    def propagate_units_into_function_args(self,
                                           token,
                                           left_token,
                                           right_token):
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])

    def propagate_units_sqrt(self, token, left_token, right_token):
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        if token.str == 'sqrt':
            unit_receiver = None
            # CASE: NAMESPACE
            if token.astParent.str == '::':
                if token.astParent.astParent:
                    unit_receiver = token.astParent.astParent
            # CASE: NAMESPACE STD
            elif token.astParent.str == '(':
                if token.astParent.astOperand2:
                    unit_receiver = token.astParent

            if not unit_receiver:
                # todo:
                return

            left_units = right_units = None
            if unit_receiver.astOperand1:
                left_units = unit_receiver.astOperand1.units
            if unit_receiver.astOperand2:
                right_units = unit_receiver.astOperand2.units
            # GET UNITS FROM INSIDE PARENS
            new_units = self.merge_units_by_set_union(left_units,
                                                      right_units)

            # DIVIDE UNITS BY TWO
            for u in new_units:
                for k, v in u.iteritems():
                    u[k] = v / 2.

            # ATTEMPT TO PROPATE UNITS ACROSS '('
            for u in new_units:
                if u not in unit_receiver.units:
                    # PROPAGATE CHILD TOKEN STATUS TO TOKEN
                    self.apply_propagation_status_to_token(token,
                                                           left_token,
                                                           right_token)
                    self.was_some_unit_changed = True
                    if self.debug_verbose:
                        s = "tw. units changed in %s" % inspect.stack()[0][3]
                        print s
                    unit_receiver.units.append(u)

    def propagate_units_inverse_trig(self, token, left_token, right_token):
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        if token.str in ['atan2', 'acos', 'asin', 'atan']:
            unit_receiver = None
            # CASE: NAMESPACE
            if token.astParent.str == '::':
                # ADD UNITS TO '::"
                unit_receiver = token.astParent.astParent
            # CASE: NAMESPACE STD
            elif token.astParent.str == '(':
                unit_receiver = token.astParent
            if not unit_receiver:
                s = 'unexpected ast value when processing inverse trig '
                s += 'linenr=%d lines=%d file=%s source_file=%s'
                s = s % (int(token.linenr),
                         len(self.source_file_lines),
                         token.file,
                         self.source_file)
                print s
                return
            for u in token.units:
                if u not in unit_receiver.units:
                    unit_receiver.units.append(u)
                    # PROPAGATE CHILD TOKEN STATUS TO TOKEN
                    self.apply_propagation_status_to_token(token,
                                                           left_token,
                                                           right_token)
                    self.was_some_unit_changed = True
                    if self.debug_verbose:
                        s = "tw. units changed in %s" % inspect.stack()[0][3]
                        print s

    def propagate_units_getXYZ(self, token, left_token, right_token):
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        if token.str in ['getX', 'getY', 'getZ']:
            unit_receiver = None
            if token.astParent.str == '(':
                unit_receiver = token.astParent
            if not unit_receiver:
                if self.debug:
                    s = 'unexpected ast value when processing getXYX linenr=%d'
                    s += ' lines=%d file=%s source_file=%s'
                    s = s % (int(token.linenr),
                             len(self.source_file_lines),
                             token.file,
                             self.source_file)
                    print s
                return
            for u in token.units:
                if u not in unit_receiver.units:
                    unit_receiver.units.append(u)
                    self.was_some_unit_changed = True
                    # THIS IS TRUE BECAUSE getX is a weaker inference
                    # - usually meters but might be m/s
                    self.is_unit_propagation_based_on_unknown_variable = True
                    token.is_unit_propagation_based_on_unknown_variable = True
                    if self.debug_verbose:
                        s = "tw. units changed in %s" % inspect.stack()[0][3]
                        print s

    def propagate_units_ternary(self, token, left_token, right_token):
        ''' TERNARY OPERATOR    x = 1 > 0 ? true_part : false_part
            '''
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        if token.str == '?':
            if not right_token or not right_token.str == ':':
                return
            self.propagate_units_across_connectors(right_token,
                                                   right_token.astOperand1,
                                                   right_token.astOperand2,
                                                   ':')

            if right_token.units:
                for u in right_token.units:
                    if u not in token.units:
                        # PROPAGATE CHILD TOKEN STATUS TO TOKEN
                        self.apply_propagation_status_to_token(token,
                                                               left_token,
                                                               right_token)
                        self.was_some_unit_changed = True
                        if self.debug_verbose:
                            s = "tw. units changed in %s"
                            s = s % inspect.stack()[0][3]
                            print s
                        token.units.append(u)

    def propagate_units_pow(self, token, left_token, right_token):
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        if token.str == 'pow':
            comma_token = None
            unit_receiver = None
            if token.astParent.str == '::':
                if token.astParent.astParent:
                    unit_receiver = token.astParent.astParent
                    comma_token = None
                    if unit_receiver.astOperand2 and \
                            unit_receiver.astOperand2.str == ',':
                        comma_token = unit_receiver.astOperand2
            # CASE: NAMESPACE STD
            elif token.astParent.str == '(':
                if token.astParent.astOperand2:
                    comma_token = token.astParent.astOperand2
                    unit_receiver = token.astParent
            if not comma_token or not unit_receiver:
                s = 'unexpected ast value when processing pow linenr=%d '
                s += 'lines=%d file=%s source_file=%s'
                s = s % (int(token.linenr),
                         len(self.source_file_lines),
                         token.file,
                         self.source_file)
                return
            if comma_token.astOperand2 and comma_token.astOperand2.isNumber:
                s = comma_token.astOperand2.str.replace('f', '')
                power_exponent = float(s)
                if comma_token.astOperand1.units:
                    # APPLY POWER TO UNITS
                    new_units = []
                    # FIRST, DEEPCOPY
                    new_units = copy.deepcopy(comma_token.astOperand1.units)
                    # NEXT, APPLY EXPONENET
                    for unit_dict in new_units:
                        for k, v in unit_dict.iteritems():
                            unit_dict[k] = power_exponent * v
                    for u in new_units:
                        if u not in unit_receiver.units:
                            # PROPAGATE CHILD TOKEN STATUS TO TOKEN
                            self.apply_propagation_status_to_token(token,
                                                                   left_token,
                                                                   right_token)
                            self.was_some_unit_changed = True
                            if self.debug_verbose:
                                s = "tw. units changed in %s"
                                s = s % inspect.stack()[0][3]
                                print s
                            unit_receiver.units.append(u)

    def propagate_units_math_min_max(self, token, left_token, right_token):
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        if token.str in ['min', 'max']:
            unit_receiver = None
            comma_token = None
            # CASE: NAMESPACE
            if token.astParent.str == '::':
                if token.astParent.astParent:
                    unit_receiver = token.astParent.astParent
                    if unit_receiver.astOperand2 and \
                            unit_receiver.astOperand2.str == ',':
                        comma_token = unit_receiver.astOperand2
            # CASE: NAMESPACE STD
            elif token.astParent.str == '(':
                if token.astParent.astOperand2:
                    if token.astParent.astOperand2.str == ',':
                        comma_token = token.astParent.astOperand2
                    unit_receiver = token.astParent
            if not unit_receiver or not comma_token:
                s = 'unexpected ast value when processing abs/fabs '
                s = 'linenr=%d lines=%d file=%s source_file=%s'
                s = s % (int(token.linenr),
                         len(self.source_file_lines),
                         token.file,
                         self.source_file)
                return
            left = right = None
            if comma_token.astOperand1:
                left = comma_token.astOperand1
            if comma_token.astOperand2:
                right = comma_token.astOperand2
            # SIDE EFFECT OF NEXT LINE SHOULD ADD UNITS TO '('
            self.propagate_units_across_connectors(comma_token,
                                                   left,
                                                   right,
                                                   ',')
            # ATTEMPT TO PROPATE UNITS ACROSS '('
            for u in comma_token.units:
                if u not in unit_receiver.units:
                    # PROPAGATE CHILD TOKEN STATUS TO TOKEN
                    self.apply_propagation_status_to_token(token,
                                                           left_token,
                                                           right_token)
                    self.was_some_unit_changed = True
                    if self.debug_verbose:
                        s = "tw. units changed in %s" % inspect.stack()[0][3]
                        print s
                    unit_receiver.units.append(u)

    def propagate_units_math_abs_fabs_floor_ceil(self,
                                                 token,
                                                 left_token,
                                                 right_token):
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        if token.str in ['abs', 'fabs', 'floor', 'ceil']:
            paren_token = None
            unit_receiver = None
            # CASE: NAMESPACE
            if token.astParent.str == '::':
                if token.astParent.astParent:
                    unit_receiver = token.astParent.astParent
                    paren_token = None
                    if unit_receiver.astOperand2 and \
                            unit_receiver.astOperand2.str == '(':
                        paren_token = unit_receiver.astOperand2
            # CASE: NAMESPACE STD
            elif token.astParent.str == '(':
                if token.astParent.astOperand2:
                    paren_token = token.astParent.astOperand2
                    unit_receiver = token.astParent

            if not paren_token or not unit_receiver:
                s = 'unexpected ast value when processing abs/fabs '
                s += 'linenr=%d lines=%d file=%s source_file=%s'
                s = s % (int(token.linenr),
                         len(self.source_file_lines),
                         token.file,
                         self.source_file)
                return

            left = right = None
            if paren_token.astOperand1:
                left = paren_token.astOperand1
            if paren_token.astOperand2:
                right = paren_token.astOperand2
            # SIDE EFFECT OF NEXT LINE SHOULD ADD UNITS TO '('
            self.propagate_units_across_connectors(paren_token,
                                                   left,
                                                   right,
                                                   '(')
            # ATTEMPT TO PROPATE UNITS ACROSS '('
            for u in paren_token.units:
                if u not in unit_receiver.units:
                    # PROPAGATE CHILD TOKEN STATUS TO TOKEN
                    self.apply_propagation_status_to_token(token,
                                                           left_token,
                                                           right_token)
                    self.was_some_unit_changed = True
                    if self.debug_verbose:
                        s = "tw. units changed in %s" % inspect.stack()[0][3]
                        print s
                    unit_receiver.units.append(u)

    def propagate_units_across_connectors(self,
                                          token,
                                          left_token,
                                          right_token,
                                          connector):
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        '''  DECORATE TREE WITH ROS MESSAGE TYPE UNITS
            input:  token  AN AST TOKEN
                    left_token  left child of token, or None
                    right_token  left child of token, or None
            returns: None  side effect on token is to modify units
        '''
        left_units = right_units = []           # INIT

        if left_token:
            left_units = left_token.units
        if right_token:
            right_units = right_token.units

        # APPLY SET UNION
        if token.str == connector:
            new_units = self.merge_units_by_set_union(left_units,
                                                      right_units)
            # UNIFY TOKENS FROM CHILDREN WITH CURRENT TOKENS
            if new_units != token.units:
                token.units = new_units
                self.was_some_unit_changed = True
                # PROPAGATE CHILD TOKEN STATUS TO TOKEN
                self.apply_propagation_status_to_token(token,
                                                       left_token,
                                                       right_token)
                if self.debug_verbose:
                    s = "tw. units changed in %s" % inspect.stack()[0][3]
                    print s

    def error_check_multiple_units(self, token, left_token, right_token):
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        if len(token.units) > 1:
            # MULTIPLE HYPOTHESIS
            if token.str == '=':
                error_string = "%s : %s  MULTIPLE UNITS BY ASSIGNMENT: %s"
                error_string = error_string % (token.file,
                                               token.linenr,
                                               token.units)
            else:
                error_string = "%s : %s  %s has MULTIPLE UNITS: %s"
                error_string = error_string % (token.file,
                                               token.linenr,
                                               token.str,
                                               token.units)
            self.handle_error(token, error_string)

    def error_check_logical_operators(self, token, left_token, right_token):
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        # NOT << and >> since theses are used in PRINT OUTS
        if token.isOp and token.str in ['&&', '||', '!']:
            error_units = ''
            if left_token and left_token.units:
                error_units += str(left_token.units)
            if right_token and right_token.units:
                error_units += str(right_token.units)
            if error_units:
                error_string = "%s : %s LOGICAL OP %s used : %s"
                error_string = error_string % (token.file,
                                               token.linenr,
                                               token.str,
                                               error_units)
                self.handle_error(token, error_string)

    def error_check_bitwise_operators(self, token, left_token, right_token):
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        # NOT << and >> since theses are used in PRINT OUTS
        if token.isOp and token.str in ['&', '|', '^', '~', '~=', '&=', '|=']:
            if left_token or right_token:
                error_units = ''
                if left_token and left_token.units:
                    error_units += str(left_token.units)
                if right_token and right_token.units:
                    error_units += str(right_token.units)
                error_string = "%s : %s BITWISE OP %s used : %s"
                error_string = error_string % (token.file,
                                               token.linenr,
                                               token.str,
                                               error_units)
                self.handle_error(token, error_string)

    def error_check_modulo(self, token, left_token, right_token):
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        # NOT << and >> since theses are used in PRINT OUTS
        if token.isOp and token.str in ['%', '%=']:
            if left_token or right_token:
                error_units = ''
                if left_token and left_token.units:
                    error_units += str(left_token.units)
                if right_token and right_token.units:
                    error_units += str(right_token.units)
                error_string = "%s : %s MODULO OP %s used : %s"
                error_string = error_string % (token.file,
                                               token.linenr,
                                               token.str,
                                               error_units)
                self.handle_error(token, error_string)

    def error_check_unitless_operators(self, token, left_token, right_token):
        ''' LOOK FOR STRONG UNITLESS TOKENS
            input:  token   cppcheck token object
                    left_token, right_token
            returns: nothing, with possible side effect of adding errors
            '''
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        if token.units:
            for token_dict in token.units:
                for value in token_dict.values():
                    # MUST HAVE BECOME UNITLESS THROUGH AN OPERATIONS
                    if value == 0:
                        error_units = str(token.units)
                        error_string = "%s : %s TOKEN IS UNITLESS %s"
                        error_string = error_string % (token.file,
                                                       token.linenr,
                                                       error_units)
                        self.handle_error(token, error_string)

    def error_check_comparison_operators(self, token, left_token, right_token):
        ''' COMPARISON OPERATORS - MUST BE THE SAME ON BOTH SIDES
            input:  token   cppcheck token object
                    left_token, right_token
            returns: nothing, with possible side effect of adding errors
            '''
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        if token.isComparisonOp:
            if left_token.units and right_token.units:  # BOTH HAVE UNITS
                if left_token.units != right_token.units:
                    error_units = str(left_token.units)
                    error_units += str(right_token.units)
                    error_string = "%s : %s COMPARISON AGAINST "
                    error_string += "DIFFERENT UNITS %s"
                    error_string = error_string % (token.file,
                                                   token.linenr,
                                                   error_units)
                    self.handle_error(token, error_string)

    def handle_error(self, token, error_string):
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        line = ''
        if self.source_file_lines:
            if str(token.file).endswith('cpp'):
                if int(token.linenr) > len(self.source_file_lines):
                    s = "token line number greater than size of source file?"
                    s += "linenr=%d lines=%d file=%s error=%s, source_file=%s"
                    s = s % (int(token.linenr),
                             len(self.source_file_lines),
                             token.file,
                             error_string,
                             self.source_file)
                    raise ValueError(s)
                line = self.source_file_lines[int(token.linenr)-1]
        error_dict = {'token': token,
                      'error_msg': error_string,
                      'line': line,
                      'file': self.current_file}
        self.errors.append(error_dict)

    def merge_units_by_set_union(self, left_units, right_units):
        ''' input: {left, right}_units - lists of unit dictionaries.
            result: set union of inputs
            '''
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        new_units = []                      # INIT
        if left_units and right_units:
            # UNITS ARE SAME - PASS THEM UP
            if left_units == right_units:
                # COPY EITHER ONE BECAUSE SAME
                new_units = copy.deepcopy(right_units)
            # BOTH RIGHT AND LEFT SIDE OF AST HAS UNITS -- GET THE 'SET'
            new_units = copy.deepcopy(left_units)
            for r in right_units:
                if r not in new_units:
                    new_units.append(copy.deepcopy(r))
        # ONLY HAVE UNITS FROM ONE SIDE OF AST TREE
        else:
            if left_units:
                new_units = copy.deepcopy(left_units)
            elif right_units:
                new_units = copy.deepcopy(right_units)
        return new_units

    def propagate_units_across_operators(self, token, left_token, right_token):
        '''  PROPAGATE UNITS ACROSS OPERATORS
            input:  AN AST TOKEN
            returns: None
        '''
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        if token.str not in ['+',
                             '-',
                             '+=',
                             '-=',
                             '*',
                             '/',
                             '*=',
                             '/=',
                             '*',
                             '*=']:
            return
        left_units = right_units = []  # INIT
        # LEFT BRANCH
        if left_token:
            left_units = left_token.units
        # RIGHT BRANCH
        if right_token:
            right_units = right_token.units
        # PROPAGATE EVEN IF WE DON'T KNOW -- ASSUME SCALAR
        if not (left_units and right_units):
            new_units = self.merge_units_by_set_union(left_units, right_units)
            # CHECK FOR DIVISION AND EMPTY LEFT BRANCH
            if token.str in ['/', '/='] and not left_units:
                # FLIP SIGNS ON NEW UNITS
                for u in new_units:
                    for k, v in u.iteritems():
                        u[k] = -1 * v
            # WEAKEN INFERENCE IF WE'RE MULTIPLYING OR
            # DIVIDING ON CONSTANTS OR UNKNOWN VARIABLES
            if token.str in ['*', '/', '*=', '/=']:
                if new_units in \
                        self.my_symbol_helper.dimensionless_units_as_lists:
                    new_units = []
                # DETECT OPERATIONS ON CONSTANTS, like x = y * 1024.23
                if left_token and not left_units:
                    if left_token.isNumber:  # TRUE FOR EITHER INT OR FLOAT
                        self.is_unit_propagation_based_on_constants = True
                        token.is_unit_propagation_based_on_constants = True
                if right_token and not right_units:
                    if right_token.isNumber:  # TRUE FOR EITHER INT OR FLOAT
                        self.is_unit_propagation_based_on_constants = True
                        token.is_unit_propagation_based_on_constants = True
                # DETECT OPERATIONS ON UNKNOWN VARIABLES
                if left_units or right_units:  # AT LEAST ONE
                    if left_token and not left_units:
                        self.is_unit_propagation_based_on_unknown_variable \
                            = True
                        token.is_unit_propagation_based_on_unknown_variable \
                            = True
                    if right_token and not right_units:
                        self.is_unit_propagation_based_on_unknown_variable \
                            = True
                        token.is_unit_propagation_based_on_unknown_variable \
                            = True

            if new_units != token.units:
                token.units = new_units
                # PROPAGATE UNCERTAINTY BASED ON CONSTANTS OR
                # UNKNOWN VARIABLES
                self.apply_propagation_status_to_token(token,
                                                       left_token,
                                                       right_token)
                self.was_some_unit_changed = True
                if self.debug_verbose:
                    s = "tw. units changed in %s" % inspect.stack()[0][3]
                    print s
            return

        new_units = []
        if token.str in ['+', '-', '+=', '-=', '*', '/', '*=', '/='] \
                and token.isOp:
            # ADDITION / SUBTRACTION - ATTEMPT TO MERGE
            if token.str in ['+', '-', '+=', '-=']:
                # ADDITION - APPLY SET UNION
                new_units = self.merge_units_by_set_union(left_units,
                                                          right_units)
            # MULTIPLICATION / DIVISION
            elif token.str in ['*', '*=', '/', '/=']:
                all_unit_dicts_from_multiplication = []
                for unit_dict_left in left_units:
                    for unit_dict_right in right_units:
                        if self.debug:
                            s = "info: calling "
                            s += "apply_multiplication_to_unit_dict: "
                            s += "%s:%s %s %d %d"
                            print s % (token.file,
                                       str(token.linenr),
                                       token.str,
                                       len(left_units),
                                       len(right_units))
                            s = "info: calling "
                            s += "apply_multiplication_to_unit_dict: "
                            s += "%s:%s %s %d %d"
                        result_units = self.apply_multiplication_to_unit_dicts(
                                unit_dict_left,
                                unit_dict_right,
                                token.str)
                        if result_units:
                            all_unit_dicts_from_multiplication.append(
                                    result_units)
                for u in all_unit_dicts_from_multiplication:
                    if u not in new_units:
                        # PROPAGATE UNCERTAINTY BASED ON
                        # CONSTANTS OR UNKNOWN VARIABLES
                        self.apply_propagation_status_to_token(token,
                                                               left_token,
                                                               right_token)
                        new_units.append(u)

        # CHECK FOR UNITLESS TOKENS AS THE RESULT OF OPERATIONS
        if self.should_check_unitless_during_multiplication:
            for unit_dict in new_units:
                for k, v in unit_dict.iteritems():
                    if v == 0:
                        # UNITLESS
                        error_string = "%s : %s UNITLESS %s" % (
                                token.file,
                                token.linenr,
                                new_units)
                        self.handle_error(token, error_string)

        # UNIFY TOKENS FROM CHILDREN WITH CURRENT TOKENS
        if new_units != token.units:
            are_all_units_zero = True  # ASSUME TRUE
            are_any_units_zero = False
            for i, u in enumerate(new_units):
                zero_test_list = [v == 0 for v in u.values()]
                are_all_units_zero = all(zero_test_list)
                are_any_units_zero = any(zero_test_list)
                # WHEN WE HAVE SOME ZERO BUT NOT ALL, FILTER THE ZERO UNITS
                if are_any_units_zero and not are_all_units_zero:
                    new_units[i] = {k: v for k, v in u.iteritems() if v != 0}
        if new_units != token.units:
            # PROPAGATE UNCERTAINTY BASED ON CONSTANTS OR UNKNOWN VARIABLES
            self.apply_propagation_status_to_token(token,
                                                   left_token,
                                                   right_token)
            token.units = new_units
            self.was_some_unit_changed = True
            if self.debug_verbose:
                s = "tw. units changed in %s" % inspect.stack()[0][3]
                print s

    def apply_propagation_status_to_token(self,
                                          token,
                                          left_token,
                                          right_token):
        ''' APPLIES PROPAGATION WEAKENING FROM CHILD TOKENS TO PARENT TOKEN
            input:  token, left_token, right_token  all tokens
            returns: none.  side effect can change variables on 'token'
            '''
        if left_token:
            if left_token.is_unit_propagation_based_on_constants:
                token.is_unit_propagation_based_on_constants = True
            if left_token.is_unit_propagation_based_on_unknown_variable:
                token.is_unit_propagation_based_on_unknown_variable = True
        if right_token:
            if right_token.is_unit_propagation_based_on_constants:
                token.is_unit_propagation_based_on_constants = True
            if right_token.is_unit_propagation_based_on_unknown_variable:
                token.is_unit_propagation_based_on_unknown_variable = True

    def apply_multiplication_to_unit_dicts(self,
                                           unit_dict_left,
                                           unit_dict_right,
                                           op):
        ''' APPLIES MULTIPLICATION AND DIVISION TO UNIT DICTIONARIES
            (BY ADDING EXPONENTS)
            input:  unit_dict_left   dictionary of units.
                ex:  {'m':1, 's':-1} is meters per second
                    unit_dict_right  same
                    op   string representing mult or div operators
            returns: new dict  with resulting units  ex: {'m':2, 's':-2}
            '''
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        return_dict = {}
        # SPECIAL HANDLING FOR RADIANS AND QUATERNIONS
        if unit_dict_left in self.my_symbol_helper.dimensionless_units \
                and unit_dict_right \
                in self.my_symbol_helper.dimensionless_units:
            # SPECIAL CASE BOTH ARE RADIANS.  CLOSED UNDER MULTIPLICATION
            return copy.deepcopy(unit_dict_left)
        if unit_dict_left not in self.my_symbol_helper.dimensionless_units \
                and unit_dict_right \
                not in self.my_symbol_helper.dimensionless_units:
            return_dict = copy.deepcopy(unit_dict_left)
        if unit_dict_left in self.my_symbol_helper.dimensionless_units:
            # DON'T PROPAGATE RADIANS
            unit_dict_left = {}
            return_dict = copy.deepcopy(unit_dict_right)
        if unit_dict_right in self.my_symbol_helper.dimensionless_units:
            # DON'T PROPAGATE RADIANS
            unit_dict_right = {}
            return_dict = copy.deepcopy(unit_dict_left)

        return_dict = copy.deepcopy(unit_dict_left)
        for unit in unit_dict_right:
            if unit in return_dict:
                # MULT WHEN BOTH DICTS HAVE SAME UNIT
                if op in ['*', '*=']:
                    # ADD OF EXPONENT IS MULT
                    return_dict[unit] += unit_dict_right[unit]
                # DIV WHEN BOTH DICTS HAVE SAME UNIT
                elif op in ['/', '/=']:
                    # SUBTRACTION OF EXPONENT IS DIV
                    return_dict[unit] -= unit_dict_right[unit]
            else:
                # ONLY ONE SIDE HAS UNIT
                if op in ['*', '*=']:
                    # COPY - THIS IS NOT A REFERNCE
                    return_dict[unit] = unit_dict_right[unit]
                elif op in ['/', '/=']:
                    return_dict[unit] = -1 * unit_dict_right[unit]  # DIVSION
        # FILTER OUT ZEROs - UNITLESS
        return_dict = {k: v for k, v in return_dict.items() if v != 0}
        return return_dict

    def collect_function_parameters_units_and_decorate_function_objects(
            self, token, left_token, right_token):
        ''' COLLECT AVAILABLE UNITS ON FUNCTION PARAMETERS FROM
            AST AND ADD THEM TO FUNCTION OBJECT
        '''
        if token.function:
            function_args_units = []
            if token.astParent.astOperand2:
                if token.astParent.astOperand2.str == ',':
                    # MULITPLE ARGS
                    # EXPECT A LIST OF LISTS
                    function_args_units = \
                            self.recurse_on_function_args(
                                    token.astParent.astOperand2)
                else:
                    # ONLY ONE ARG
                    if token.astParent.astOperand2.units:
                        # LIST OF LIST
                        function_args_units = \
                                [token.astParent.astOperand2.units]
            if function_args_units:
                if len(function_args_units) != len(token.function.arg_units):
                    if self.debug:
                        print token.file
                        s = 'line %s   '
                        s += 'len(function_args_units)'
                        s += '== %d len(token.function.arg_units) = %d'
                        print s % (token.linenr,
                                   len(function_args_units),
                                   len(token.function.arg_units))
                    return
                for i, u in enumerate(function_args_units):
                    new_dict = {'linenr': int(token.linenr),
                                'units': u,
                                'functionId': token.function.Id,
                                'token': token}
                    if new_dict not in token.function.arg_units[i]:
                        token.function.arg_units[i].append(new_dict)

    def recurse_on_function_args(self, comma_token):
        ''' RECURSIVELY COLLECT UNITS FOR FUNCTION ARGUMENTS
            input: ccpcheck token object str=','
            output: list of units in arg order
            '''
        my_return_list = []
        if comma_token.astOperand1:
            if comma_token.astOperand1.str == ',':
                left_branch_units_list =  \
                    self.recurse_on_function_args(comma_token.astOperand1)
                if left_branch_units_list and \
                        isinstance(left_branch_units_list[0], list):
                    for u in left_branch_units_list:
                        my_return_list.append(u)
                else:
                    my_return_list.append(left_branch_units_list)
            else:
                my_return_list.append(comma_token.astOperand1.units)
        if comma_token.astOperand2:
            my_return_list.append(comma_token.astOperand2.units)
        return my_return_list

    def debug_walk(self, token):
        ''' RUDIMENTARY TERMINAL PRINT HOTWASH DEBUG
        '''
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])

        if token.isRoot:
            print '- '*42 + ' ROOT'
            if self.source_file_lines:
                print "lines yes",
                if self.source_file in token.file:
                    print 'source in token'
                    print self.source_file_lines[int(token.linenr)-1],
                else:
                    print 'self.source_file not in token.file'

        self.recursive_debug_walk(token, 0)

    def recursive_debug_walk(self, token, indent):
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        if not token:
            return

        print '%s %s   %s             ' % (token.linenr,
                                           "-" * indent,
                                           token.str),
        if token.variable:
            print self.my_symbol_helper.find_variable_type(token.variable),
            print self.my_symbol_helper.find_compound_variable_name_for_variable_token(token),
        if token.units:
            print token.units,
        if token.function:
            print token.function.argument,
            print 'ret: ' + str(token.function.return_units),
        if token.is_unit_propagation_based_on_constants:
            print ' C ',
        if token.is_unit_propagation_based_on_unknown_variable:
            print ' UV ',
        print
        self.recursive_debug_walk(token.astOperand1, indent + 1)  # LEFT
        self.recursive_debug_walk(token.astOperand2, indent + 1)  # RIGHT
        if token.link:
            print '%s %s   %s' % (token.link.linenr,
                                  "+" * indent,
                                  token.link.str)

    def find_first_variable(self, token, left_token, right_token):
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        ''' ASSUME THIS WILL BE CALLED ON LHS OF ASSIGNMENT.
            SHOULD ONLY BE ONE VARIABLE
        '''
        if token.variable:
            self.return_value_list.append(token.variable)

    def recurse_and_collect_string(self, AST_token):
        if self.debug_verbose:
            self.cps_unit_checker.handle_debug_verbose(inspect.stack()[0][3])
        """ SIMPLE RECURSION FOR LEFT-HAND SIDE OF ASSIGNMENT STATMENTS
        """
        # PROTECT FROM NULL
        if not AST_token:
            return ''
        my_return_string = ''
        # LEFT RECURSE
        if AST_token.astOperand1:
            my_return_string +=  \
                self.recurse_and_collect_string(AST_token.astOperand1)
        # SELF
        my_return_string += AST_token.str
        # RIGHT RECURSE
        if AST_token.astOperand2:
            my_return_string +=  \
                self.recurse_and_collect_string(AST_token.astOperand2)

        return my_return_string



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import git
import os
import sys
from subprocess import Popen
from database_connector import DatabaseConnector
from cps_units_checker import CPSUnitsChecker
import time

import click


@click.command()
@click.argument('file_containing_list_of_files_to_be_analyzed') 
#TODO: make this generic or read from the same location as phriky units
@click.option('--include_dir', default='', help='include directory for cppcheck')
# @click.option('--debug_print_ast/--no-debug_print_ast', default=False, help='Very verbose debug print of Abstract Syntax Trees and unit decorations')
@click.option('--show_high_confidence/--no_show_high_confidence', default=True, help='Should show hig-confidence inconsistences. Defaults to True')
@click.option('--show_low_confidence/--no_show_low_confidence', default=False, help='Should show low-confidence inconsistencies. Defaults to False.')
@click.option('--use_var_name_heuristic/--no_use_var_name_heuristic', default=False, help='Should use variable name heuristics to assignm units. Experimental.')
@click.option('--check_unit_smell/--no_check_unit_smell', default=False, help='Should check for strange unit assignments (unit smells). See error_checker.py')
@click.option('--delete_dump_file/--no_delete_dump_file', default=False, help='Will not delete temporary cppcheck "dump" file.')
@click.option('--use_existing_dump_file/--no_use_existing_dump_file', default=True, help='Reuses existing dump file, if possible.')
@click.option('--only_find_files_with_units/--no_only_find_files_with_units', default=False, help='No inconsistency analysis. Only finds files with physical units.')
@click.option('--write_results_to_database/--no_write_results_to_database', default=True, help='Write results to database.  See DatabaseConnector.py for details.')
@click.option('--note_for_database', default='', help='note for database analysis record')
def main(file_containing_list_of_files_to_be_analyzed, 
        include_dir, 
        # debug_print_ast,
        show_high_confidence,
        show_low_confidence,
        use_var_name_heuristic,
        check_unit_smell,
        delete_dump_file,
        use_existing_dump_file,
        only_find_files_with_units,
        write_results_to_database,
        note_for_database):
    """Console script for phriky_units"""
    # 'Detect physical unit inconsistencies in C++ code, especially ROS code.')

    # VERIFY TARGET FILE EXISTS
    if not os.path.exists(file_containing_list_of_files_to_be_analyzed):
        print ('File not found: %s' % file_containing_list_of_files_to_be_analyzed)

    database_analysis_audit_id = 0
    if write_results_to_database:
        # database_analysis_audit_id = 157
        dbc = DatabaseConnector()
        dbc.VERSION = CPSUnitsChecker().VERSION
        # INSERT nOTE
        note = note_for_database
        if not note:
            note = file_containing_list_of_files_to_be_analyzed
        dbc.create_database_entry_for_one_time_analysis(file_containing_list_of_files_to_be_analyzed, note)
        database_analysis_audit_id = dbc.database_analysis_audit_id
        print( 'NEW DATABASE ANALYSIS: ID=%d' % database_analysis_audit_id)

    args = ['./phriky_units.py']
    if write_results_to_database and database_analysis_audit_id > 0:
        args.append('--write-results-to-database')
        args.append('--database-analysis-audit-id='+str(database_analysis_audit_id))
    if check_unit_smell:
        args.append('--check-unit-smell')
    if use_existing_dump_file:
        args.append('--use-existing-dump-file')
    # if delete_dump_file:
        # args.append('--delete-dump-file')
    # else:
        # args.append('--no-delete-dump-file')

    if use_var_name_heuristic:
        args.append('--use-var-name-heuristic')
    

    # if include_dir:
        # args.append('--include_dir ' + include_dir)

    all_files_to_analyze = [f.strip() for f in open(file_containing_list_of_files_to_be_analyzed).readlines()]
    iter_of_all_files_to_analyze = iter(all_files_to_analyze)


    # for file_name in all_files_to_analyze:
        # if not os.path.exists(file_name):
            # print ('Could not find file: %s' % file_name)
            # continue
        # current_repository = ''
        # # GET THE GIT REPOSITORY
        # if write_results_to_database:
            # s = git.cmd.Git(os.path.dirname(file_name)).execute(['git', 'rev-parse', '--show-toplevel'])  # same as above
            # current_repository = os.path.basename(s)
        # # OPEN PHRIKY-UNITS
        # phriky_units_process = Popen(' '.join(args + ['--current_repository='+current_repository]+ [file_name]),  shell=True)
        # phriky_units_process.communicate()
        # # processes_started += 1
        # # list_of_processes.append(phriky_units_process)

    # ------------------------------------------
    # RUN MULTIPLE PROCESS AT ONCE
    # ------------------------------------------
    number_of_processes = 3
    # number_of_processes = 5
    list_of_processes = []
    first = True
    processes_started = 0
    processes_finished = 0

    process_counter = 0

    while ( len(list_of_processes) > 0 )  or first:
        first = False
        # print len(list_of_processes)
        # print ('Number of processes active: %d' % len(list_of_processes))
        # ADD PROCESSES UNTIL LIMIT REACHED
        while len(list_of_processes) < number_of_processes:
            file_name = next(iter_of_all_files_to_analyze, None)
            if not file_name:
                break
            # TEST FOR NON-ASCII CHARACTERS IN NAME
            try:
                file_name.decode('ascii')
            except UnicodeDecodeError:
                print("Skipping non-ascii filename %s" % file_name)
                continue
            # TEST FOR FILE EXISTS
            if not os.path.exists(file_name):
                print ('Could not find file: %s' % file_name)
                continue
            current_repository = ''
            # GET THE GIT REPOSITORY
            if write_results_to_database:
                s = git.cmd.Git(os.path.dirname(file_name)).execute(['git', 'rev-parse', '--show-toplevel'])  # same as above
                current_repository = os.path.basename(s)
            # OPEN PHRIKY-UNITS
            # phriky_units_process = Popen(' '.join(args + ['--current_repository='+current_repository]+ [file_name]),  shell=False)
            phriky_units_process = Popen(' '.join(args + ['--current-repository='+current_repository]+ [file_name]),  shell=True)
            processes_started += 1
            # print("starting proc %d : %s " % (phriky_units_process.pid, file_name))
            list_of_processes.append(phriky_units_process)

        p_to_remove = None
        # CHECK IF THE PROCESS IS DONE, AND REMOVE IF POSSIBLE
        # print list_of_processes
        for p in list_of_processes:
            # ASK PROCESS TO UPDATE ITS RETURNCODE
            # if process_counter == 100:
                # print ("running : %d " %  p.pid)
            if not p.poll() :
                if p.returncode is not None:
                    # p.terminate()
                    # print("process %d complete returncode: %s" % (p.pid, str(p.returncode)))
                    # p.communicate()
                    p_to_remove = p
                    break
                # else:
                    # print("process %d still running p.poll() is None, returncode is None: " % p.pid)
                    # pass
            # else:
                # print("process %d still running: p.poll() is not None " % p.pid)
                # pass

        if process_counter == 100:
            process_counter = 0
        process_counter += 1


        if p_to_remove and p_to_remove in list_of_processes:
            # p_to_remove.terminate()
            # print("removing process %d: " % p_to_remove.pid)
            list_of_processes.remove(p_to_remove)
            # p_to_remove = None
            processes_finished += 1
        # GO EASY ON YOUR OS
        # print ('S-F: %d ' % (processes_started - processes_finished))
        time.sleep(0.05)

    # for file_name in iter(all_files_to_analyze):







if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pymysql
from unit_error_types import UnitErrorTypes

class DatabaseConnector:
    """Adds MYSQL connectivity and includes query functions."""

    def __init__(self):
        self.static_unit_id_lookup = {}
        self.init_units()
        self.database_host = 'localhost'
        self.database_user='robot'
        self.database_name =  'database_units_checking_analysis'
        self.database_password ='RobotRobot1!'
        self.db_analysis_audit_row_id  = 0
        self.debug = False
        self.VERSION = ''
        self.current_repository = ''
        self.current_file_under_analysis = ''
    
    def init_units(self):
        self.static_unit_id_lookup = {
            'amp': 12,
            'degree_360': 9,
            'degree_celsius':10,
            'kilogram': 4,
            'lux':8,
            'meter': 1,
            'pascal': 6,
            'quaternion':5,
            'radian':3,
            'second':2,
            'tesla':7 ,
            'volt':11
            }

    def execute_this_database_query(self, query_as_string):
        ''' TAKES SQL QUERY AS INPUT AND RUNS IT
            input: query_as_string    SQL query as string
            output: none
            '''
        if self.debug:
            print query_as_string
        conn = pymysql.connect(host=self.database_host,
                             user=self.database_user,
                             password=self.database_password,
                             db=self.database_name)
        c = conn.cursor()
	c.execute(query_as_string)
        lastrowid = c.lastrowid
        self.db_most_recent_insert_id = lastrowid
	conn.commit()
        # SAVE LAST INSERT ID - USEFUL FOR INSERTS IN 1-TO-MANY-TABLES
	conn.close()
        return lastrowid


    def add_class_and_units_to_all_list(self, class_name, units):
        ''' WHEN TRACKING ALL UNIT TYPES, ADD THEM TO DATA STRUCTURE HERE
            input:  class_name str the name of the ROS class or str (ex: 'atan2') that merits units
                    units : the dictionary or units
            output: None, side effect updates local data structures.
            '''
        if not class_name in self.all_classes_and_units_for_this_file_as_dict:
            self.all_classes_and_units_for_this_file_as_dict[class_name] = units
        else:
            pass  # WE DON'T CASE BECAUSE WE'RE ONLY TRACKING UNIQUE UNITS, NOT COUNTS

    def insert_file_unit_class_records(self):
        ''' INSERTS UNIT RECORDS FOR ALL FILES INTO DATABASE
            input: None - Requires self.all_classes_and_units_for_this_file_as_dict
            output: None - side effect: updates database
            '''
        if not self.all_classes_and_units_for_this_file_as_dict:
            return
        query_string = ''' INSERT INTO file_and_units
                    (date_time, file_uri)
                    VALUES
                    (NOW(),'''
        query_string += "'" + str(self.current_file_under_analysis) + "'"
        query_string += ')'
        self.execute_this_database_query(query_string)
        self.file_and_unit_id = self.db_most_recent_insert_id
        for class_name, units in self.all_classes_and_units_for_this_file_as_dict.iteritems():
            query_string = ''' INSERT INTO units_and_classes
                            (
                            file_and_units_id
                            , ros_header_class
                            , units
                            ) VALUES (
                            ''' + str(self.file_and_unit_id)
            query_string += ", '" + class_name + "'"
            query_string += ', "' + str(units) + '"'
            query_string += ')'
            self.execute_this_database_query(query_string)


    def create_database_entry_for_one_time_analysis(self, input_file='', notes=''):
        '''UPDATES LOCAL DATABASE FOR ONE ANALYSIS - USUALLY TESTING
            '''
        query_string = ''' INSERT INTO analysis_audit
            (datetime, version, input_file, notes)
            VALUES
            (NOW(),'''
        query_string += "'" + str(self.VERSION) + "',"
        query_string += "'" + str(input_file) + "',"
        query_string += "'" + str(notes) + "'"
        query_string += ')'
        self.execute_this_database_query(query_string)
        self.database_analysis_audit_id = self.db_most_recent_insert_id

    def update_database(self, list_of_unit_error_objects=None):
        ''' UPDATES THE LOCAL DATABASE WITH A SUMMARY OF THE RESULTS OF THE ANALYSIS OF THIS FILE
            requires: the self.db_analysis_audit_row_id is correct
            input:  data_record  list of numbers of each kind of error ordered by unit_error_type [1, 2, 3, 4, 5, 6, 7] 
                    list_of_unit_error_objects   UnitError Objects 
            returns: None
            side-effect(s): new records in the local database
            '''
        query_string = '''INSERT INTO file_analysis
		(
		analysis_audit_id,
		datetime,
		repository,
		file_name,
		some_ROS_units_found
		)
		VALUES
		(''' + str(self.database_analysis_audit_id) + ''',
		NOW(),''' 
        query_string += "'" + self.current_repository + "'," 
        query_string += "'" + self.current_file_under_analysis + "'," 
        query_string += str(int(self.found_ros_units_in_this_file)) + ')'
        # query_string += ','.join([str(x) for x in data_record]) + ')'
        # query_string += '''%d,%d,%d,%d,%d,%d,%d)''' % data_record
		# + '''?,?,?,?,?,?,?)''' % data_record
        file_analysis_id = self.execute_this_database_query(query_string)
        # NOW INSERT ERROR DETAILS FOR EACH ERROR
        if list_of_unit_error_objects:
            for e in list_of_unit_error_objects:
                query_string_error_details = '''INSERT INTO error_details
                (
                file_analysis_id,
                linenr,
                token,
                var_name,
                unit_error_types_id,
                file_uri,
                is_unit_propagation_based_on_constants,
                is_unit_propagation_based_on_unknown_variable,
                is_warning
                ) 
                VALUES 
                (''' + str(file_analysis_id)
                query_string_error_details += ',' + str(e.linenr) + ','
                if e.token:
                    query_string_error_details += "'" + e.token.str + "',"
                else:
                    query_string_error_details += "'',"
                query_string_error_details += "'" + e.var_name + "',"
                query_string_error_details += "'" + str(e.ERROR_TYPE) + "',"
                query_string_error_details += "'" + str(e.get_file_URI_where_error_occured()) + "',"
                query_string_error_details += "'" + str(int(e.is_unit_propagation_based_on_constants)) + "',"
                query_string_error_details += "'" + str(int(e.is_unit_propagation_based_on_unknown_variable)) + "',"
                query_string_error_details += "'" + str(int(e.is_warning)) + "')"
                # print query_string_error_details
                error_details_id = self.execute_this_database_query(query_string_error_details)

                if e.ERROR_TYPE in [UnitErrorTypes().VARIABLE_MULTIPLE_UNITS, UnitErrorTypes().UNIT_SMELL]:
                    # UPDATE DATABASE WITH UNITS OF FIRST ASSIGNMENT AND UNITS AT PROBLEM LOCATION
                    variable_insert_query = '''INSERT INTO variable  
                        (
                        file_analysis_id,
                        error_details_id,
                        var_name
                        ) VALUES (
                        ''' + str(file_analysis_id) + ','
                    variable_insert_query += str(error_details_id) + ','
                    variable_insert_query += "'" + e.var_name + "'" + ')'
                    variable_id = self.execute_this_database_query(variable_insert_query)
                    for unit_assignment_linenr, dict_of_units_and_tokens in e.all_units_assigned_to_var_as_dict.iteritems():
                        for unit_dict in dict_of_units_and_tokens['units']:  # THIS IS A LIST

                            # INSERT RECORD FOR UNIT ASSIGNMENT
                            variable_unit_assignment_insert_query = '''INSERT INTO variable_unit_assignment
                                (
                                variable_id,
                                linenr
                                ) VALUES (''' + str(variable_id) + ','
                            variable_unit_assignment_insert_query += str(unit_assignment_linenr) + ')'
                            variable_unit_assignment_id = self.execute_this_database_query(variable_unit_assignment_insert_query)
                            # print 'UUUU:%s ' % str(unit_dict)
                            for unit, power in unit_dict.iteritems():
                                unit_assignment_insert_query = ''' INSERT INTO unit_assignment
                                    (
                                    variable_unit_assignment_id,
                                    unit_types_id,
                                    power
                                    ) VALUES ('''
                                unit_assignment_insert_query += str(variable_unit_assignment_id) + ','
                                unit_assignment_insert_query += str(self.static_unit_id_lookup[unit]) + ','
                                unit_assignment_insert_query += str(power)
                                unit_assignment_insert_query += ')'
                                # print unit_assignment_insert_query
                                self.execute_this_database_query(unit_assignment_insert_query)





# @mainpage cppcheckdata
#
# @brief This is a Python module that helps you access Cppcheck dump data.
#
# License: No restrictions, use this as you need.<br><br>
#

import xml.etree.ElementTree as ET
import argparse

# Directive class. Contains information about each preprocessor directive
# in the source code.
#
# file and linenr denote the location where the directive is defined.
#
# To iterate through all directives use such code:
# @code
# data = cppcheckdata.parsedump(...)
# for cfg in data.configurations:
#   for directive in cfg.directives:
#     print(directive.str)
# @endcode
#


class Directive:
    # The directive line, with all C or C++ comments removed
    str = None
    # name of (possibly included) file where directive is defined
    file = None
    # line number in (possibly included) file where directive is defined
    linenr = None

    def __init__(self, element):
        self.str = element.get('str')
        self.file = element.get('file')
        self.linenr = element.get('linenr')


# Token class. Contains information about each token in the source code.
#
# The CppcheckData.tokenlist is a list of Token items
#
# C++ class: http://cppcheck.sourceforge.net/devinfo/doxyoutput/classToken.html
#
# To iterate through all tokens use such code:
# @code
# data = cppcheckdata.parsedump(...)
# for cfg in data.configurations:
#   code = ''
#   for token in cfg.tokenlist:
#     code = code + token.str + ' '
#   print(code)
# @endcode
#


class Token:
    Id = None
    # Token string
    str = None
    # Next token in tokenlist. For last token, next is None.
    next = None
    # Previous token in tokenlist. For first token, previous is None,
    previous = None
    linkId = None
    # Linked token in tokenlist. Each '(', '[' and '{' are linked to the
    # corresponding '}', ']' and ')'. For templates, the '<' is linked to
    # the corresponding '>'.
    link = None
    scopeId = None
    # Scope information for this token. See the Scope class.
    scope = None
    # Is this token a symbol name
    isName = False
    # Is this token a number, for example 123, 12.34
    isNumber = False
    # Is this token a int value such as 1234
    isInt = False
    # Is this token a int value such as 12.34
    isFloat = False
    # Is this token a string literal such as "hello"
    isString = False
    # string length for string literal
    strlen = None
    # Is this token a char literal such as 'x'
    isChar = False
    # Is this token a operator
    isOp = False
    # Is this token a arithmetic operator
    isArithmeticalOp = False
    # Is this token a assignment operator
    isAssignmentOp = False
    # Is this token a comparison operator
    isComparisonOp = False
    # Is this token a logical operator: && ||
    isLogicalOp = False
    # varId for token, each variable has a unique non-zero id
    varId = None
    variableId = None
    # Variable information for this token. See the Variable class.
    #
    # Example code:
    # @code
    # data = cppcheckdata.parsedump(...)
    # code = ''
    # for token in data.tokenlist:
    #   code = code + token.str
    #   if token.variable:
    #     if token.variable.isLocal:
    #       code = code + ':localvar'
    #     if token.variable.isArgument:
    #       code = code + ':arg'
    #   code = code + ' '
    # print(code)
    # @endcode
    variable = None
    functionId = None
    # If this token points at a function call, this attribute has the Function
    # information. See the Function class.
    function = None
    valuesId = None
    # Possible values of token
    #
    # Example code:
    # @code
    # data = cppcheckdata.parsedump(...)
    # code = ''
    # for token in data.tokenlist:
    #   code = code + token.str
    #   if token.values:
    #     # print values..
    #     code = code + '{'
    #     for value in token.values:
    #       if value.intvalue:
    #         code = code + str(value.intvalue) + ' '
    #     code = code + '}'
    #   code = code + ' '
    # print(code)
    # @endcode
    values = None

    typeScopeId = None
    # type scope (token->type()->classScope)
    typeScope = None

    astParentId = None
    # syntax tree parent
    astParent = None
    astOperand1Id = None
    # syntax tree operand1
    #
    # Example code:
    # @code
    # data = cppcheckdata.parsedump(...)
    # for token in data.tokenlist:
    #
    #   # is this a addition?
    #   if token.str == '+':
    #
    #     # print LHS operand
    #     print(token.astOperand1.str)
    #
    # @endcode
    astOperand1 = None
    astOperand2Id = None
    # syntax tree operand2
    #
    # Example code:
    # @code
    # data = cppcheckdata.parsedump(...)
    # for token in data.tokenlist:
    #
    #   # is this a division?
    #   if token.str == '/':
    #
    #     # print RHS operand
    #     print(token.astOperand2.str)
    #
    # @endcode
    astOperand2 = None

    # file name
    file = None
    # line number
    linenr = None

    def __init__(self, element):
        self.Id = element.get('id')
        self.str = element.get('str')
        self.next = None
        self.previous = None
        self.scopeId = element.get('scope')
        self.scope = None
        type = element.get('type')
        if type == 'name':
            self.isName = True
        elif type == 'number':
            self.isNumber = True
            if element.get('isInt'):
                self.isInt = True
            elif element.get('isFloat'):
                self.isFloat = True
        elif type == 'string':
            self.isString = True
            self.strlen = int(element.get('strlen'))
        elif type == 'char':
            self.isChar = True
        elif type == 'op':
            self.isOp = True
            if element.get('isArithmeticalOp'):
                self.isArithmeticalOp = True
            elif element.get('isAssignmentOp'):
                self.isAssignmentOp = True
            elif element.get('isComparisonOp'):
                self.isComparisonOp = True
            elif element.get('isLogicalOp'):
                self.isLogicalOp = True
        if element.get('unitType'):
            self.unitType = element.get('unitType')
        self.linkId = element.get('link')
        self.link = None
        self.varId = element.get('varId')
        self.variableId = element.get('variable')
        self.variable = None
        self.functionId = element.get('function')
        self.function = None
        self.valuesId = element.get('values')
        self.values = None
        self.typeScopeId = element.get('type-scope')
        self.typeScope = None
        self.astParentId = element.get('astParent')
        self.astParent = None
        self.astOperand1Id = element.get('astOperand1')
        self.astOperand1 = None
        self.astOperand2Id = element.get('astOperand2')
        self.astOperand2 = None
        self.file = element.get('file')
        self.linenr = element.get('linenr')

    def setId(self, IdMap):
        self.scope = IdMap[self.scopeId]
        self.link = IdMap[self.linkId]
        self.variable = IdMap[self.variableId]
        self.function = IdMap[self.functionId]
        self.values = IdMap[self.valuesId]
        # if self.typeScopeId is not '0x0':
        if self.typeScopeId in IdMap:
            self.typeScope = IdMap[self.typeScopeId]
        self.astParent = IdMap[self.astParentId]
        self.astOperand1 = IdMap[self.astOperand1Id]
        self.astOperand2 = IdMap[self.astOperand2Id]

    # Get value if it exists
    # Returns None if it doesn't exist
    def getValue(self, v):
        if not self.values:
            return None
        for value in self.values:
            if value.intvalue == v:
                return value
        return None

# Scope. Information about global scope, function scopes,
# class scopes, inner scopes, etc.
# C++ class:
# http://cppcheck.sourceforge.net/devinfo/doxyoutput/classScope.html


class Scope:
    Id = None
    classStartId = None

    # The { Token for this scope
    classStart = None
    classEndId = None
    # The } Token for this scope
    classEnd = None
    # Name of this scope.
    # For a function scope, this is the function name;
    # for a class scope, this is the class name.
    className = None
    # Type of scope: Global, Function, Class, If, While
    type = None

    def __init__(self, element):
        self.Id = element.get('id')
        self.className = element.get('className')
        self.classStartId = element.get('classStart')
        self.classStart = None
        self.classEndId = element.get('classEnd')
        self.classEnd = None
        self.nestedInId = element.get('nestedId')
        self.nestedIn = None
        self.type = element.get('type')
        self.functionId = element.get('function')

    def setId(self, IdMap):
        self.classStart = IdMap[self.classStartId]
        self.classEnd = IdMap[self.classEndId]
        self.nestedIn = IdMap[self.nestedInId]
        self.function = IdMap[self.functionId]
        if self.nestedIn:
            self.foo = self.nestedInId
        else:
            self.bar = self.nestedInId

# Information about a function
# C++ class:
# http://cppcheck.sourceforge.net/devinfo/doxyoutput/classFunction.html


class Function:
    Id = None
    argument = None
    argumentId = None
    tokenDef = None
    tokenDefId = None
    name = None

    def __init__(self, element):
        self.Id = element.get('id')
        self.tokenDefId = element.get('tokenDef')
        self.name = element.get('name')
        self.argument = {}
        self.argumentId = {}
        for arg in element:
            self.argumentId[arg.get('nr')] = arg.get('id')

    def setId(self, IdMap):
        for argnr, argid in self.argumentId.items():
            self.argument[argnr] = IdMap[argid]
        self.tokenDef = IdMap[self.tokenDefId]

# Information about a variable
# C++ class:
# http://cppcheck.sourceforge.net/devinfo/doxyoutput/classVariable.html


class Variable:
    Id = None
    nameTokenId = None
    # name token in variable declaration
    nameToken = None
    typeStartTokenId = None
    # start token of variable declaration
    typeStartToken = None
    typeEndTokenId = None
    # end token of variable declaration
    typeEndToken = None
    # Is this variable a function argument?
    isArgument = False
    # Is this variable an array?
    isArray = False
    # Is this variable a class or struct?
    isClass = False
    # Is this variable a local variable?
    isLocal = False
    # Is this variable a pointer
    isPointer = False
    # Is this variable a reference
    isReference = False
    # Is this variable static?
    isStatic = False

    def __init__(self, element):
        self.Id = element.get('id')
        self.nameTokenId = element.get('nameToken')
        self.nameToken = None
        self.typeStartTokenId = element.get('typeStartToken')
        self.typeStartToken = None
        self.typeEndTokenId = element.get('typeEndToken')
        self.typeEndToken = None
        self.isArgument = element.get('isArgument') == 'true'
        self.isArray = element.get('isArray') == 'true'
        self.isClass = element.get('isClass') == 'true'
        self.isLocal = element.get('isLocal') == 'true'
        self.isPointer = element.get('isPointer') == 'true'
        self.isReference = element.get('isReference') == 'true'
        self.isStatic = element.get('isStatic') == 'true'

    def setId(self, IdMap):
        self.nameToken = IdMap[self.nameTokenId]
        self.typeStartToken = IdMap[self.typeStartTokenId]
        self.typeEndToken = IdMap[self.typeEndTokenId]

# ValueFlow class


class ValueFlow:
    # ValueFlow::Value class
    # Each possible value has a ValueFlow::Value item.
    # Each ValueFlow::Value either has a intvalue or tokvalue
    # C++ class:
    # http://cppcheck.sourceforge.net/
    #      devinfo/doxyoutput/classValueFlow_1_1Value.html

    class Value:
        # integer value
        intvalue = None
        # token value
        tokvalue = None
        # condition where this Value comes from
        condition = None

        def __init__(self, element):
            self.intvalue = element.get('intvalue')
            if self.intvalue:
                self.intvalue = int(self.intvalue)
            self.tokvalue = element.get('tokvalue')
            self.condition = element.get('condition-line')
            if self.condition:
                self.condition = int(self.condition)

    Id = None

    # Possible values
    values = None

    def __init__(self, element):
        self.Id = element.get('id')
        self.values = []
        for value in element:
            self.values.append(ValueFlow.Value(value))

# Configuration class
# This class contains the directives, tokens, scopes, functions,
# variables and value flows for one configuration.


class Configuration:
    # Name of the configuration, "" for default
    name = ''
    # List of Directive items
    directives = []
    # List of Token items
    tokenlist = []
    # List of Scope items
    scopes = []
    # List of Function items
    functions = []
    # List of Variable items
    variables = []
    # List of ValueFlow values
    valueflow = []

    def __init__(self, confignode):
        self.name = confignode.get('cfg')
        self.directives = []
        self.tokenlist = []
        self.scopes = []
        self.functions = []
        self.variables = []
        self.valueflow = []

        for element in confignode:
            if element.tag == 'directivelist':
                for directive in element:
                    self.directives.append(Directive(directive))

            if element.tag == 'tokenlist':
                for token in element:
                    self.tokenlist.append(Token(token))

                # set next/previous..
                prev = None
                for token in self.tokenlist:
                    token.previous = prev
                    if prev:
                        prev.next = token
                    prev = token
            if element.tag == 'scopes':
                for scope in element:
                    self.scopes.append(Scope(scope))
                    for functionList in scope:
                        if functionList.tag == 'functionList':
                            for function in functionList:
                                self.functions.append(Function(function))
            if element.tag == 'variables':
                for variable in element:
                    self.variables.append(Variable(variable))
            if element.tag == 'valueflow':
                for values in element:
                    self.valueflow.append(ValueFlow(values))

        IdMap = {}
        IdMap[None] = None
        IdMap['0'] = None
        for token in self.tokenlist:
            IdMap[token.Id] = token
        for scope in self.scopes:
            IdMap[scope.Id] = scope
        for function in self.functions:
            IdMap[function.Id] = function
        for variable in self.variables:
            IdMap[variable.Id] = variable
        for values in self.valueflow:
            IdMap[values.Id] = values.values

        for token in self.tokenlist:
            token.setId(IdMap)
        for scope in self.scopes:
            scope.setId(IdMap)
        for function in self.functions:
            function.setId(IdMap)
        for variable in self.variables:
            variable.setId(IdMap)

# Class that makes cppcheck dump data available
# Contains a list of Configuration instances
#
# To iterate through all configurations use such code:
# @code
# data = cppcheckdata.parsedump(...)
# for cfg in data.configurations:
#     print('cfg: ' + cfg.name)
# @endcode
#
# To iterate through all tokens in each configuration use such code:
# @code
# data = cppcheckdata.parsedump(...)
# for cfg in data.configurations:
#     print('cfg: ' + cfg.name)
#     code = ''
#         for token in cfg.tokenlist:
#             code = code + token.str + ' '
#     print('    ' + code)
# @endcode
#
# To iterate through all scopes (functions, types, etc) use such code:
# @code
# data = cppcheckdata.parsedump(...)
# for cfg in data.configurations:
#     print('cfg: ' + cfg.name)
#     for scope in cfg.scopes:
#         print('    type:' + scope.type + ' name:' + scope.className)
# @endcode
#
#


class CppcheckData:
    # List of Configurations
    configurations = []

    def __init__(self, filename):
        self.configurations = []

        data = ET.parse(filename)
        # root is 'dumps' node, each config has its own 'dump' subnode.
        for cfgnode in data.getroot():
            self.configurations.append(Configuration(cfgnode))

# parse a cppcheck dump file


def parsedump(filename):
    return CppcheckData(filename)

# Check if type of ast node is float/double


def astIsFloat(token):
    if not token:
        return False
    if token.str == '.':
        return astIsFloat(token.astOperand2)
    if '+-*/%'.find(token.str) == 0:
        if astIsFloat(token.astOperand1):
            return True
        return astIsFloat(token.astOperand2)
    if not token.variable:
        # float literal?
        if token.str[0].isdigit():
            for c in token.str:
                if c == 'f' or c == '.' or c == 'E':
                    return True
        return False
    typeToken = token.variable.typeStartToken
    endToken = token.variable.typeEndToken
    while typeToken != endToken:
        if typeToken.str == 'float' or typeToken.str == 'double':
            return True
        typeToken = typeToken.next
    if typeToken.str == 'float' or typeToken.str == 'double':
        return True
    return False


class CppCheckFormatter(argparse.HelpFormatter):
    '''
        Properly formats multiline argument helps
    '''
    def _split_lines(self, text, width):
        # this is the RawTextHelpFormatter._split_lines
        if text.startswith('R|'):
            return text[2:].splitlines()
        return argparse.HelpFormatter._split_lines(self, text, width)


def ArgumentParser():
    '''
        Returns an argparse argument parser with an already-added
        argument definition for -t/--template
    '''
    parser = argparse.ArgumentParser(formatter_class=CppCheckFormatter)
    parser.add_argument('-t', '--template', metavar='<text>',
                        default='{callstack}: ({severity}) {message}',
                        help="R|Format the error messages. E.g.\n"
                        "'{file}:{line},{severity},{id},{message}' or\n"
                        "'{file}({line}):({severity}) {message}' or\n"
                        "'{callstack} {message}'\n"
                        "Pre-defined templates: gcc, vs, edit")
    return parser


def reportError(template, callstack=[], severity='', message='', id=''):
    '''
        Format an error message according to the template.

        :param template: format string, or 'gcc', 'vs' or 'edit'.
        :param callstack: e.g. [['file1.cpp',10],['file2.h','20'], ... ]
        :param severity: e.g. 'error', 'warning' ...
        :param id: message ID.
        :param message: message text.
    '''
    # expand predefined templates
    if template == 'gcc':
        template = '{file}:{line}: {severity}: {message}'
    elif template == 'vs':
        template = '{file}({line}): {severity}: {message}'
    elif template == 'edit':
        template = '{file} +{line}: {severity}: {message}'
    # compute 'callstack}, {file} and {line} replacements
    stack = ' -> '.join(['['+f+':'+str(l)+']' for (f, l) in callstack])
    file = callstack[-1][0]
    line = str(callstack[-1][1])
    # format message
    return template.format(callstack=stack, file=file, line=line,
                           severity=severity, message=message, id=id)
